###### Trainer ##########
import warnings
from typing import Optional, Union
from collections import defaultdict
import os
import time
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

import copy
import math
import numpy as np
import random
import scanpy as sc
from anndata import concat

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.data import Data
from torch_geometric.data import InMemoryDataset
from torch_geometric.sampler import NeighborSampler
from torch_geometric.loader import NeighborLoader, GraphSAINTEdgeSampler, NodeLoader, DataLoader
import torch_geometric.transforms as T

import matplotlib
from matplotlib import pyplot as plt
from ..data.dataprocessors import prepare_data
from ..data.dataloaders import initialize_dataloaders
from .utils import _cycle_iterable, print_progress, EarlyStopping
from .metrics import eval_metrics
from .basetrainermixin import BaseMixin
# from .transfer_anno import weighted_knn_trainer, weighted_knn_transfer

# from torch.optim.lr_scheduler import StepLR
# from scheduler import CosineAnnealingWarmRestarts
# GAMMA=0.9
warnings.filterwarnings("ignore", category=UserWarning, module="torch_geometric")

class GarfieldTrainer(BaseMixin):
    """
    Garfield model trainer.
    """
    def __init__(self,
                 adata,
                 model: nn.Module,
                 label_name,
                 used_pca_feat,
                 adj_key,
                 edge_val_ratio,
                 edge_test_ratio,
                 node_val_ratio,
                 node_test_ratio,
                 augment_type,
                 num_neighbors,
                 loaders_n_hops,
                 edge_batch_size,
                 node_batch_size,
                 reload_best_model,
                 use_early_stopping,
                 early_stopping_kwargs,
                 monitor,
                 verbose,
                 seed,
                 **kwargs):
        """
        Initializes the GarfieldTrainer class, which handles data preparation, model initialization,
        and training of the Garfield model.

        Parameters
        ----------
        adata : AnnData
            Annotated data matrix.
        model : nn.Module
            The Garfield model to be trained.
        label_name : str
            Column name for labels in the annotated data.
        used_pca_feat : Bool
            Whether used pca features or not for node feature.
        adj_key : str
            Key for the adjacency matrix (e.g., spatial connectivity) in the data.
        edge_val_ratio : float
            Proportion of edges to use for validation in edge-level tasks.
        edge_test_ratio : float
            Proportion of edges to use for testing in edge-level tasks.
        node_val_ratio : float
            Proportion of nodes to use for validation in node-level tasks.
        node_test_ratio : float
            Proportion of nodes to use for testing in node-level tasks.
        augment_type : str
            Type of data augmentation to apply (e.g., 'dropout', 'svd').
        num_neighbors : int
            Number of neighbors to sample for each node in graph-based tasks.
        loaders_n_hops : int
            Number of hops to consider for neighbors in graph-based tasks.
        edge_batch_size : int
            Batch size for edge-level tasks.
        node_batch_size : int or None
            Batch size for node-level tasks. If None, it will be determined automatically.
        reload_best_model : bool
            Whether to reload the best model after training.
        use_early_stopping : bool
            Whether to apply early stopping during training.
        early_stopping_kwargs : dict
            Additional arguments for early stopping (e.g., patience, delta).
        monitor : bool
            Whether to print monitoring logs during training.
        verbose : bool
            Whether to print detailed logs during training.
        seed : int
            Seed for random number generation to ensure reproducibility.
        kwargs : dict
            Additional arguments for training configuration.
        """
        self.adata = adata
        self.model = model
        self.label_name_ = label_name
        self.used_pca_feat_ = used_pca_feat
        self.adj_key_ = adj_key # spatial_connectivities

        ## data split
        self.edge_val_ratio_ = edge_val_ratio # 0.1
        self.edge_train_ratio_ = 1 - edge_val_ratio
        self.edge_test_ratio_ = edge_test_ratio # 0.
        self.node_val_ratio_ = node_val_ratio
        self.node_train_ratio_ = 1 - node_val_ratio
        self.node_test_ratio_ = node_test_ratio

        ## data process
        self.augment_type_ = augment_type

        ## data loader
        # self.loader_type_ = self.args.loader_type
        self.n_sampled_neighbors_ = num_neighbors
        self.loaders_n_hops_ = loaders_n_hops
        self.edge_batch_size_ = edge_batch_size # 512
        self.node_batch_size_ = node_batch_size # None,

        ## other parameters
        self.reload_best_model_ = reload_best_model
        self.use_early_stopping_ = use_early_stopping
        self.early_stopping_kwargs_ = early_stopping_kwargs
        self.monitor_ = monitor
        self.verbose_ = verbose
        self.loaders_n_hops_ = kwargs.pop("loaders_n_hops", 1)
        self.gradient_clipping_ = kwargs.pop("gradient_clipping", 0.)
        self.epoch = -1
        self.training_time = 0
        self.optimizer = None
        self.best_epoch = None
        self.best_model_state_dict = None
        self.early_stopping_kwargs_ = (self.early_stopping_kwargs_ if
                                       self.early_stopping_kwargs_ else {})
        if not "early_stopping_metric" in self.early_stopping_kwargs_:
            if self.edge_val_ratio_ > 0 and self.node_val_ratio_ > 0:
                self.early_stopping_kwargs_["early_stopping_metric"] = (
                    "val_global_loss")
            else:
                self.early_stopping_kwargs_["early_stopping_metric"] = (
                    "train_global_loss")
        self.early_stopping = EarlyStopping(**self.early_stopping_kwargs_)

        print("\n--- INITIALIZING TRAINER ---")
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        # seed = 2024
        random.seed(seed)
        os.environ['PYTHONHASHSEED'] = str(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        self.model.to(self.device)

        # Load dataset
        # Prepare data and get node-level and edge-level training and validation
        # splits
        data_dict = prepare_data(
            adata=self.adata,
            label_name=self.label_name_,
            used_pca_feat=self.used_pca_feat_,
            adj_key=self.adj_key_,
            edge_val_ratio=self.edge_val_ratio_,
            edge_test_ratio=self.edge_test_ratio_,
            node_val_ratio=self.node_val_ratio_,
            node_test_ratio=self.node_test_ratio_,)
        self.node_masked_data = data_dict["node_masked_data"]
        self.edge_train_data = data_dict["edge_train_data"]
        self.edge_val_data = data_dict["edge_val_data"]
        self.n_nodes_train = self.node_masked_data.train_mask.sum().item()
        self.n_nodes_val = self.node_masked_data.val_mask.sum().item()
        self.n_edges_train = self.edge_train_data.edge_label_index.size(1)
        self.n_edges_val = self.edge_val_data.edge_label_index.size(1)
        print(f"Number of training nodes: {self.n_nodes_train}")
        print(f"Number of validation nodes: {self.n_nodes_val}")
        print(f"Number of training edges: {self.n_edges_train}")
        print(f"Number of validation edges: {self.n_edges_val}")

        # Determine node batch size automatically if not specified
        if self.node_batch_size_ is None:
            self.node_batch_size_ = int(self.edge_batch_size_ / math.floor(
                self.n_edges_train / self.n_nodes_train))
        print(f"Edge batch size: {self.edge_batch_size_}")
        print(f"Node batch size: {self.node_batch_size_}")

        # Initialize node-level and edge-level dataloaders
        loader_dict = initialize_dataloaders(
            node_masked_data=self.node_masked_data,
            edge_train_data=self.edge_train_data,
            edge_val_data=self.edge_val_data,
            edge_batch_size=self.edge_batch_size_,
            node_batch_size=self.node_batch_size_,
            n_direct_neighbors=self.n_sampled_neighbors_,
            n_hops=self.loaders_n_hops_,
            edges_directed=False,
            neg_edge_sampling_ratio=1.)
        self.edge_train_loader = loader_dict["edge_train_loader"]
        self.edge_val_loader = loader_dict.pop("edge_val_loader", None)
        self.node_train_loader = loader_dict["node_train_loader"]
        self.node_val_loader = loader_dict.pop("node_val_loader", None)

    def train(self,
              n_epochs,
              n_epochs_no_edge_recon, # : int=0
              learning_rate,
              weight_decay,
              gradient_clipping,
              lambda_edge_recon,
              lambda_gene_expr_recon,
              lambda_latent_contrastive_instanceloss,
              lambda_latent_contrastive_clusterloss,
              lambda_omics_recon_mmd_loss
              ):
        """
        Trains the Garfield model for a specified number of epochs with joint edge-level and node-level tasks.

        Parameters
        ----------
        n_epochs : int
            Number of epochs to train the model.
        n_epochs_no_edge_recon : int
            Number of epochs to train without edge reconstruction.
        learning_rate : float
            Learning rate for the optimizer.
        weight_decay : float
            Weight decay (L2 regularization) for the optimizer.
        gradient_clipping : float
            Maximum value for gradient clipping to prevent gradient explosion.
        lambda_edge_recon : float
            Weight for edge reconstruction loss.
        lambda_gene_expr_recon : float
            Weight for gene expression reconstruction loss.
        lambda_latent_contrastive_instanceloss : float
            Weight for instance-level latent contrastive loss.
        lambda_latent_contrastive_clusterloss : float
            Weight for cluster-level latent contrastive loss.
        lambda_omics_recon_mmd_loss : float
            Weight for omics reconstruction MMD loss.

        Returns
        -------
        None
        """
        self.n_epochs_ = n_epochs
        self.n_epochs_no_edge_recon_ = n_epochs_no_edge_recon
        self.lr_ = learning_rate
        self.weight_decay_ = weight_decay
        self.gradient_clipping_ = gradient_clipping
        self.lambda_edge_recon_ = lambda_edge_recon
        self.lambda_gene_expr_recon_ = lambda_gene_expr_recon
        self.lambda_latent_contrastive_instanceloss = lambda_latent_contrastive_instanceloss
        self.lambda_latent_contrastive_clusterloss = lambda_latent_contrastive_clusterloss
        self.lambda_omics_recon_mmd_loss_ = lambda_omics_recon_mmd_loss

        print("\n--- MODEL TRAINING ---")
        start_time = time.time()
        self.epoch_logs = defaultdict(list)
        self.model.train()
        params = filter(lambda p: p.requires_grad, self.model.parameters())
        self.optimizer = torch.optim.Adam(params,
                                          lr=learning_rate,
                                          weight_decay=weight_decay)

        for self.epoch in range(n_epochs):
            if self.epoch < self.n_epochs_no_edge_recon_:
                self.edge_recon_active = False
            else:
                self.edge_recon_active = True

            self.iter_logs = defaultdict(list)
            self.iter_logs["n_train_iter"] = 0
            self.iter_logs["n_val_iter"] = 0

            # Jointly loop through edge- and node-level batches, repeating node-
            # level batches until edge-level batches are complete
            for edge_train_data_batch, node_train_data_batch in zip(
                    self.edge_train_loader,
                    _cycle_iterable(self.node_train_loader)):  # itertools.cycle
                # resulted in memory leak
                # Forward pass node-level batch
                node_train_data_batch = node_train_data_batch.to(self.device)
                node_train_model_output = self.model(
                    data_batch=node_train_data_batch,
                    decoder_type="omics",
                    augment_type=self.augment_type_)

                # Forward pass edge-level batch
                edge_train_data_batch = edge_train_data_batch.to(self.device)
                edge_train_model_output = self.model(
                    data_batch=edge_train_data_batch,
                    decoder_type="graph",
                    augment_type=None)

                # Calculate training loss
                train_loss_dict = self.model.loss(
                    edge_model_output=edge_train_model_output,
                    node_model_output=node_train_model_output,
                    lambda_edge_recon=lambda_edge_recon,
                    lambda_gene_expr_recon=lambda_gene_expr_recon,
                    lambda_latent_contrastive_instanceloss=lambda_latent_contrastive_instanceloss,
                    lambda_latent_contrastive_clusterloss=lambda_latent_contrastive_clusterloss,
                    lambda_omics_recon_mmd_loss=lambda_omics_recon_mmd_loss
                )

                train_global_loss = train_loss_dict["global_loss"]
                train_optim_loss = train_loss_dict["optim_loss"]

                if self.verbose_:
                    for key, value in train_loss_dict.items():
                        self.iter_logs[f"train_{key}"].append(value.item())
                else:
                    self.iter_logs["train_global_loss"].append(
                        train_global_loss.item())
                    self.iter_logs["train_optim_loss"].append(
                        train_optim_loss.item())
                self.iter_logs["n_train_iter"] += 1
                # Optimize for training loss
                self.optimizer.zero_grad()

                train_optim_loss.backward()
                # Clip gradients
                if self.gradient_clipping_ > 0:
                    torch.nn.utils.clip_grad_value_(self.model.parameters(), self.gradient_clipping_)
                self.optimizer.step()

            # Validate model
            if (self.edge_val_loader is not None and
                    self.node_val_loader is not None):
                self.test_metrics_epoch(
                    lambda_edge_recon=lambda_edge_recon,
                    lambda_gene_expr_recon=lambda_gene_expr_recon,
                    lambda_latent_contrastive_instanceloss=lambda_latent_contrastive_instanceloss,
                    lambda_latent_contrastive_clusterloss=lambda_latent_contrastive_clusterloss,
                    lambda_omics_recon_mmd_loss=lambda_omics_recon_mmd_loss)
            elif (self.edge_val_loader is None and
                  self.node_val_loader is not None):
                warnings.warn("You have specified a node validation set but no "
                              "edge validation set. Skipping validation...")
            elif (self.edge_val_loader is not None and
                  self.node_val_loader is None):
                warnings.warn("You have specified an edge validation set but no"
                              " node validation set. Skipping validation...")

            # Convert iteration level logs into epoch level logs
            for key in self.iter_logs:
                if key.startswith("train"):
                    self.epoch_logs[key].append(
                        np.array(self.iter_logs[key]).sum() /
                        self.iter_logs["n_train_iter"])
                if key.startswith("val"):
                    self.epoch_logs[key].append(
                        np.array(self.iter_logs[key]).sum() /
                        self.iter_logs["n_val_iter"])

            # Monitor epoch level logs
            if self.monitor_:
                print_progress(self.epoch, self.epoch_logs, self.n_epochs_)

            # Check early stopping
            if self.use_early_stopping_:
                if self.is_early_stopping():
                    break

        # Track training time and load best model
        self.training_time += (time.time() - start_time)
        minutes, seconds = divmod(self.training_time, 60)
        print(f"Model training finished after {int(minutes)} min {int(seconds)}"
              " sec.")
        if self.best_model_state_dict is not None and self.reload_best_model_:
            print("Using best model state, which was in epoch "
                  f"{self.best_epoch + 1}.")
            self.model.load_state_dict(self.best_model_state_dict)

        self.model.eval()
        # Calculate after training validation metrics
        if self.edge_val_loader is not None:
            self.validate_end()

    @torch.no_grad()
    def test_metrics_epoch(self,
                           lambda_edge_recon,
                           lambda_gene_expr_recon,
                           lambda_latent_contrastive_instanceloss,
                           lambda_latent_contrastive_clusterloss,
                           lambda_omics_recon_mmd_loss
                           ):
        """
        Evaluates the Garfield model at the end of each epoch on validation data.

        Parameters
        ----------
        lambda_edge_recon : float
            Weight for edge reconstruction loss.
        lambda_gene_expr_recon : float
            Weight for gene expression reconstruction loss.
        lambda_latent_contrastive_instanceloss : float
            Weight for instance-level latent contrastive loss.
        lambda_latent_contrastive_clusterloss : float
            Weight for cluster-level latent contrastive loss.
        lambda_omics_recon_mmd_loss : float
            Weight for omics reconstruction MMD loss.

        Returns
        -------
        None
        """
        self.model.eval()

        edge_recon_probs_val_accumulated = np.array([])
        edge_recon_labels_val_accumulated = np.array([])

        # Jointly loop through edge- and node-level batches, repeating node-
        # level batches until edge-level batches are complete
        for edge_val_data_batch, node_val_data_batch in zip(
                self.edge_val_loader, _cycle_iterable(self.node_val_loader)):
            # Forward pass node level batch
            node_val_data_batch = node_val_data_batch.to(self.device)
            node_val_model_output = self.model(
                data_batch=node_val_data_batch,
                decoder_type="omics",
                augment_type=self.augment_type_)

            # Forward pass edge level batch
            edge_val_data_batch = edge_val_data_batch.to(self.device)
            edge_val_model_output = self.model(
                data_batch=edge_val_data_batch,
                decoder_type="graph",
                augment_type=None)

            # Calculate validation loss
            val_loss_dict = self.model.loss(
                edge_model_output=edge_val_model_output,
                node_model_output=node_val_model_output,
                lambda_edge_recon=lambda_edge_recon,
                lambda_gene_expr_recon=lambda_gene_expr_recon,
                lambda_latent_contrastive_instanceloss=lambda_latent_contrastive_instanceloss,
                lambda_latent_contrastive_clusterloss=lambda_latent_contrastive_clusterloss,
                lambda_omics_recon_mmd_loss=lambda_omics_recon_mmd_loss
            )

            val_global_loss = val_loss_dict["global_loss"]
            val_optim_loss = val_loss_dict["optim_loss"]
            if self.verbose_:
                for key, value in val_loss_dict.items():
                    self.iter_logs[f"val_{key}"].append(value.item())
            else:
                self.iter_logs["val_global_loss"].append(val_global_loss.item())
                self.iter_logs["val_optim_loss"].append(val_optim_loss.item())
            self.iter_logs["n_val_iter"] += 1

            # Calculate evaluation metrics
            edge_recon_probs_val = torch.sigmoid(
                edge_val_model_output["edge_recon_logits"])
            edge_recon_labels_val = edge_val_model_output["edge_recon_labels"]
            edge_recon_probs_val_accumulated = np.append(
                edge_recon_probs_val_accumulated,
                edge_recon_probs_val.detach().cpu().numpy())
            edge_recon_labels_val_accumulated = np.append(
                edge_recon_labels_val_accumulated,
                edge_recon_labels_val.detach().cpu().numpy())
        val_eval_dict = eval_metrics(
            edge_recon_probs=edge_recon_probs_val_accumulated,
            edge_labels=edge_recon_labels_val_accumulated)
        if self.verbose_:
            self.epoch_logs["val_auroc_score"].append(
                val_eval_dict["auroc_score"])
            self.epoch_logs["val_auprc_score"].append(
                val_eval_dict["auprc_score"])
            self.epoch_logs["val_best_acc_score"].append(
                val_eval_dict["best_acc_score"])
            self.epoch_logs["val_best_f1_score"].append(
                val_eval_dict["best_f1_score"])

        self.model.train()

    @torch.no_grad()
    def validate_end(self):
        """
        Evaluates the Garfield model after training on the validation dataset.

        Returns
        -------
        None
        """
        self.model.eval()

        # Get edge-level ground truth and predictions
        edge_recon_probs_val_accumulated = np.array([])
        edge_recon_labels_val_accumulated = np.array([])
        for edge_val_data_batch in self.edge_val_loader:
            edge_val_data_batch = edge_val_data_batch.to(self.device)

            edge_val_model_output = self.model(
                data_batch=edge_val_data_batch,
                decoder_type="graph",
                augment_type=None)

            # Calculate evaluation metrics
            edge_recon_probs_val = torch.sigmoid(
                edge_val_model_output["edge_recon_logits"])
            edge_recon_labels_val = edge_val_model_output["edge_recon_labels"]
            edge_recon_probs_val_accumulated = np.append(
                edge_recon_probs_val_accumulated,
                edge_recon_probs_val.detach().cpu().numpy())
            edge_recon_labels_val_accumulated = np.append(
                edge_recon_labels_val_accumulated,
                edge_recon_labels_val.detach().cpu().numpy())

        # Get node-level ground truth and predictions
        omics_pred_dict_val_accumulated = np.array([])
        omics_truth_dict_val_accumulated = np.array([])
        for node_val_data_batch in self.node_val_loader:
            node_val_data_batch = node_val_data_batch.to(self.device)

            node_val_model_output = self.model(
                data_batch=node_val_data_batch,
                decoder_type="omics",
                augment_type=self.augment_type_)

            omics_truth_dict_val = node_val_model_output["truth_x"]
            omics_pred_dict_val = node_val_model_output["recon_features"]
            omics_pred_dict_val_accumulated = np.append(
                omics_pred_dict_val_accumulated,
                omics_pred_dict_val.detach().cpu().numpy())
            omics_truth_dict_val_accumulated = np.append(
                omics_truth_dict_val_accumulated,
                omics_truth_dict_val.detach().cpu().numpy())

        val_eval_dict = eval_metrics(
            edge_recon_probs=edge_recon_probs_val_accumulated,
            edge_labels=edge_recon_labels_val_accumulated,
            omics_recon_pred=omics_pred_dict_val_accumulated,
            omics_recon_truth=omics_truth_dict_val_accumulated)
        print("\n--- MODEL EVALUATION ---")
        print(f"val AUROC score: {val_eval_dict['auroc_score']:.4f}")
        print(f"val AUPRC score: {val_eval_dict['auprc_score']:.4f}")
        print(f"val best accuracy score: {val_eval_dict['best_acc_score']:.4f}")
        print(f"val best F1 score: {val_eval_dict['best_f1_score']:.4f}")
        print(f"val MSE score: {val_eval_dict['gene_expr_mse_score']:.4f}")

    def is_early_stopping(self) -> bool:
        """
        Check whether to apply early stopping, update learning rate and save
        best model state.

        Returns
        ----------
        stop_training:
            If `True`, stop NicheCompass model training.
        """
        early_stopping_metric = self.early_stopping.early_stopping_metric
        current_metric = self.epoch_logs[early_stopping_metric][-1]
        if self.early_stopping.update_state(current_metric):
            self.best_model_state_dict = copy.deepcopy(self.model.state_dict())
            self.best_epoch = self.epoch

        continue_training, reduce_lr = self.early_stopping.step(current_metric)
        if reduce_lr:
            for param_group in self.optimizer.param_groups:
                param_group["lr"] *= self.early_stopping.lr_factor
            print(f"New learning rate is {param_group['lr']}.\n")
        stop_training = not continue_training
        return stop_training

    @classmethod
    def _get_init_params_from_dict(cls, dct):
        init_params = {
            # Input options
            'data_dir': dct['data_dir_'],
            'project_name': dct['project_name_'],
            'adata_list': dct['adata_list_'],
            'profile': dct['profile_'],
            'data_type': dct['data_type_'],
            'sub_data_type': dct['sub_data_type_'],
            'weight': dct['weight_'],
            'graph_const_method': dct['graph_const_method_'],
            'genome': dct['genome_'],
            'sample_col': dct['sample_col_'],

            ## whether to use metacell mode
            'metacell': dct['metacell_'],
            'metacell_size': dct['metacell_size_'],
            'single_n_top_genes': dct['single_n_top_genes_'],
            'n_pcs': dct['n_pcs_'],

            # Preprocessing options
            'filter_cells_rna': dct['filter_cells_rna_'],
            'min_features': dct['min_features_'],
            'min_cells': dct['min_cells_'],
            'keep_mt': dct['keep_mt_'],
            'normalize': dct['normalize_'],
            'target_sum': dct['target_sum_'],
            'used_hvg': dct['used_hvg_'],
            'used_scale': dct['used_scale_'],
            'used_feat': dct['used_feat_'],
            'rna_n_top_features': dct['rna_n_top_features_'],
            'atac_n_top_features': dct['atac_n_top_features_'],
            'n_components': dct['n_components_'],
            'n_neighbors': dct['n_neighbors_'],
            'svd_solver': dct['svd_solver_'],
            'method': dct['method_'],
            'metric': dct['metric_'],
            'resolution_tol': dct['resolution_tol_'],
            'leiden_runs': dct['leiden_runs_'],
            'leiden_seed': dct['leiden_seed_'],
            'verbose': dct['verbose_'],

            # Model options
            'augment_type': dct['augment_type_'],
            'svd_q': dct['svd_q_'],
            'gnn_layer': dct['gnn_layer_'],
            'conv_type': dct['conv_type_'],
            'hidden_dims': dct['hidden_dims_'],
            'bottle_neck_neurons': dct['bottle_neck_neurons_'],
            'cluster_num': dct['cluster_num_'],
            'num_heads': dct['num_heads_'],
            'concat': dct['concat_'],
            'used_edge_weight': dct['used_edge_weight_'],
            'used_recon_exp': dct['used_recon_exp_'],
            'used_DSBN': dct['used_DSBN_'],
            'used_mmd': dct['used_mmd_'],
            'test_split': dct['test_split_'],
            'val_split': dct['val_split_'],
            'batch_size': dct['batch_size_'],
            'loader_type': dct['loader_type_'],
            'num_neighbors': dct['num_neighbors_'],
            'epochs': dct['epochs_'],
            'dropout': dct['dropout_'],
            'mmd_temperature': dct['mmd_temperature_'],
            'instance_temperature': dct['instance_temperature_'],
            'cluster_temperature': dct['cluster_temperature_'],
            'l2_reg': dct['l2_reg_'],
            'patience': dct['patience_'],
            'monitor_only_val_losses': dct['monitor_only_val_losses_'],
            'gradient_clipping': dct['gradient_clipping_'],
            'learning_rate': dct['learning_rate_'],
            'weight_decay': dct['weight_decay_'],

            # Other options
            'projection': dct['projection_'],
            'impute': dct['impute_'],
            'outdir': dct['outdir_'],
            'load': False
        }

        return init_params

    ## Loss curve
    def plot_loss_curves(self,
                         title="Losses Curve"
                         ) -> plt.figure:
        """
        Plot loss curves.

        Parameters
        ----------
        loss_dict:
            Dictionary containing the training and validation losses.

        Returns
        ----------
        fig:
            Matplotlib figure of loss curves.
        """
        # Plot epochs as integers
        ax = plt.figure().gca()
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))

        # Plot loss
        train_global_loss = self.iter_logs["train_global_loss"]
        train_optim_loss = self.iter_logs["train_optim_loss"]
        val_global_loss = self.iter_logs["val_global_loss"]
        val_optim_loss = self.iter_logs["val_optim_loss"]
        # x = np.linspace(0, len(train_global_loss), num=len(train_global_loss))
        plt.plot(train_global_loss, label="Train_global_loss")
        plt.plot(train_optim_loss, label="Train_optim_loss")
        plt.plot(val_global_loss, label="Val_global_loss")
        plt.plot(val_optim_loss, label="Val_optim_loss")
        # for loss_key, loss in self.iter_logs.items():
        #     plt.plot(loss, label=loss_key)
        plt.title(title)
        plt.ylabel("loss")
        plt.xlabel("epoch")
        plt.legend() # loc="upper right"

        # Retrieve figure
        fig = plt.gcf()
        plt.close()
        return fig






"""
This module contains the decoder used by the Garfield model.
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.data import Data
from torch_geometric.nn import GCNConv, GATConv, GATv2Conv
from .utils import DSBatchNorm, compute_cosine_similarity

class GATDecoder(nn.Module):
    def __init__(self, in_channels, hidden_dims, out_channels, conv_type, num_heads, dropout, concat,
                 num_domains='', used_edge_weight=False, used_DSBN=False):
        """
        Initializes the GATDecoder, which consists of multiple Graph Attention Network (GAT) layers
        followed by domain-specific normalization (if applicable).

        Parameters
        ----------
        in_channels : int
            Number of input feature dimensions.
        hidden_dims : list[int]
            List of output dimensions for each hidden GAT layer, in ascending order.
        out_channels : int
            Number of output feature dimensions.
        conv_type : str
            Type of GAT convolution layer to use ('GAT' or 'GATv2Conv').
        num_heads : int
            Number of attention heads for each GAT layer.
        dropout : float
            Dropout rate applied to GAT layers.
        concat : bool
            Whether to concatenate the output of all attention heads or not.
        num_domains : int or str
            Number of domains for domain-specific batch normalization (DSBN). If `1`, regular batch normalization is used.
        used_edge_weight : bool, optional
            Whether to use edge weights in the GAT layers. Default is False.
        used_DSBN : bool, optional
            Whether to use domain-specific batch normalization (DSBN). Default is False.
        """
        super(GATDecoder, self).__init__()
        self.used_DSBN = used_DSBN
        self.used_edge_weight = used_edge_weight
        self.layers = nn.ModuleList()
        self.norm = nn.ModuleList()
        self.dropout = dropout
        if conv_type == 'GAT':
            GATLayer = GATConv
        elif conv_type == 'GATv2Conv':
            GATLayer = GATv2Conv

        num_hidden_layers = len(hidden_dims)
        num_heads_list = [num_heads] * num_hidden_layers
        concat_list = [concat] * num_hidden_layers
        for i in range(num_hidden_layers):
            if concat_list[i]:
                current_dim = hidden_dims[::-1][i] * num_heads_list[i]
            else:
                current_dim = hidden_dims[::-1][i]  # [::-1] 代表反转

            if type(num_domains) == int:
                if num_domains == 1:  # TO DO
                    norm = nn.BatchNorm1d(current_dim)
                else:
                    # num_domains >1 represent domain-specific batch normalization of n domain
                    norm = DSBatchNorm(current_dim, num_domains)
            else:
                norm = None
            self.norm.append(norm)

        current_dim = in_channels  # in_channels
        dropout_list = [dropout] * num_hidden_layers
        for i in range(num_hidden_layers):
            layer = GATLayer(
                in_channels=current_dim,
                out_channels=hidden_dims[::-1][i],
                heads=num_heads_list[i],
                dropout=dropout_list[i],
                concat=concat_list[i],
                edge_dim=1 if self.used_edge_weight else None
            )
            self.layers.append(layer)
            if concat_list[i]:
                current_dim = hidden_dims[::-1][i] * num_heads_list[i]
            else:
                current_dim = hidden_dims[::-1][i]

        ### 数据集总重构的layers
        self.conv_recon = GATLayer(in_channels=current_dim,
                                   out_channels=out_channels,
                                   heads=num_heads,
                                   concat=False,
                                   edge_dim=1 if self.used_edge_weight else None,
                                   dropout=dropout)

    def forward(self, x, data):
        """
        Performs a forward pass through the GAT decoder layers and reconstructs the node features.

        Parameters
        ----------
        x : torch.Tensor
            Node features (shape: [num_nodes, feature_dim]).
        data : Data
            PyTorch Geometric Data object containing edge index, edge attributes, and domain labels.

        Returns
        -------
        torch.Tensor
            Reconstructed node features after passing through the GAT layers and applying normalization (if applicable).
        """
        edge_index, y, edge_index_all = data.edge_index, data.y, data.edge_attr
        edge_weight = edge_index_all[:, 2]
        # edge_weight = torch.ones(edge_index.shape[1]).cpu().numpy()  # 先将张量转移到CPU，再转为numpy

        for idx, layer in enumerate(self.layers):
            x, _ = layer(x, edge_index, edge_attr=edge_weight if self.used_edge_weight else None,
                         return_attention_weights=True)
            if self.used_DSBN:
                if self.norm:
                    if len(x) == 1:
                        pass
                    elif self.norm[0].__class__.__name__ == 'DSBatchNorm':
                        x = self.norm[idx](x, y)
                    else:
                        x = self.norm[idx](x)
                    x = F.relu(x)

        recon_x, _ = self.conv_recon(x, edge_index,
                                     edge_attr=edge_weight if self.used_edge_weight else None,
                                     return_attention_weights=True)

        return recon_x


### GCN decoder
class GCNDecoder(nn.Module):
    def __init__(self, in_channels, hidden_dims, out_channels, dropout=0.2,
                 num_domains='', used_edge_weight=False, used_DSBN=False):
        """
        Initializes the GCNDecoder, consisting of multiple Graph Convolutional Network (GCN) layers followed by domain-specific normalization (if applicable).

        Parameters
        ----------
        in_channels : int
            Number of input feature dimensions.
        hidden_dims : list[int]
            List of output dimensions for each hidden GCN layer, in ascending order.
        out_channels : int
            Number of output feature dimensions.
        dropout : float, optional
            Dropout rate applied to GCN layers. Default is 0.2.
        num_domains : int or str
            Number of domains for domain-specific batch normalization (DSBN). If `1`, regular batch normalization is used.
        used_edge_weight : bool, optional
            Whether to use edge weights in the GCN layers. Default is False.
        used_DSBN : bool, optional
            Whether to use domain-specific batch normalization (DSBN). Default is False.
        """
        super(GCNDecoder, self).__init__()
        # 如果 hidden_channels 是单一数字，将其转换成单元素列表
        if isinstance(hidden_dims, int):
            hidden_dims = [hidden_dims]

        self.used_DSBN = used_DSBN
        self.used_edge_weight = used_edge_weight
        self.layers = nn.ModuleList()
        self.norm = nn.ModuleList()

        # 创建一个包含所有GCN层的列表
        gcn_layers = []
        hidden_dims = hidden_dims[::-1]  # 反转
        total_layers = [in_channels] + hidden_dims
        for i in range(len(total_layers) - 1):
            gcn_layers.append(GCNConv(total_layers[i], total_layers[i + 1], dropout=dropout))

        # 使用 nn.ModuleList 以确保所有层都被正确注册
        self.gcn_layers = nn.ModuleList(gcn_layers)

        # 输出层
        self.gcn_recon = GCNConv(hidden_dims[-1], out_channels, dropout=dropout)

        # self.norm 层
        num_hidden_layers = len(hidden_dims)
        for i in range(num_hidden_layers):
            current_dim = hidden_dims[i]

            if type(num_domains) == int:
                if num_domains == 1:  # TO DO
                    norm = nn.BatchNorm1d(current_dim)
                else:
                    norm = DSBatchNorm(current_dim,
                                       num_domains)  # num_domains >1 represent domain-specific batch normalization of n domain
            else:
                norm = None
            self.norm.append(norm)

    def forward(self, x, data):
        """
        Performs a forward pass through the GCN decoder layers and reconstructs the node features.

        Parameters
        ----------
        x : torch.Tensor
            Node features (shape: [num_nodes, feature_dim]).
        data : Data
            PyTorch Geometric Data object containing edge index, edge attributes, and domain labels.

        Returns
        -------
        torch.Tensor
            Reconstructed node features after passing through the GCN layers and applying normalization (if applicable).
        """
        edge_index, y, edge_index_all = data.edge_index, data.y, data.edge_attr
        edge_weight = edge_index_all[:, 2]
        # edge_weight = torch.ones(edge_index.shape[1]).cpu().numpy()  # 先将张量转移到CPU，再转为numpy

        ### latent
        for idx, layer in enumerate(self.gcn_layers):
            x = layer(x, edge_index,
                      edge_weight=edge_weight if self.used_edge_weight else None)

            if self.used_DSBN:
                if self.norm:
                    if len(x) == 1:
                        pass
                    elif self.norm[0].__class__.__name__ == 'DSBatchNorm':
                        print('Perform DSBN normalization...')
                        x = self.norm[idx](x, y)
                    else:
                        print('Perform batch normalization...')
                        x = self.norm[idx](x)
                    x = F.relu(x)

        x_recon = self.gcn_recon(x, edge_index,
                                 edge_weight=edge_weight if self.used_edge_weight else None)

        return x_recon


class CosineSimGraphDecoder(nn.Module):
    """
    Cosine similarity graph decoder class.

    Takes the concatenated latent feature vectors z of the source and
    target nodes as input, and calculates the element-wise cosine similarity
    between source and target nodes to return the reconstructed edge logits.

    The sigmoid activation function to compute reconstructed edge probabilities
    is integrated into the binary cross entropy loss for computational
    efficiency.

    Parameters
    ----------
    dropout_rate:
        Probability of nodes to be dropped during training.
    """

    def __init__(self,
                 dropout_rate: float = 0.):
        super().__init__()
        print("COSINE SIM GRAPH DECODER -> "
              f"dropout_rate: {dropout_rate}")

        self.dropout = nn.Dropout(dropout_rate)

    def forward(self,
                z: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the cosine similarity graph decoder.

        Parameters
        ----------
        z:
            Concatenated latent feature vector of the source and target nodes
            (dim: 4 * edge_batch_size x n_gps due to negative edges).

        Returns
        ----------
        edge_recon_logits:
            Reconstructed edge logits (dim: 2 * edge_batch_size due to negative
            edges).
        """
        z = self.dropout(z)

        # Compute element-wise cosine similarity
        edge_recon_logits = compute_cosine_similarity(
            z[:int(z.shape[0] / 2)],  # ´edge_label_index[0]´
            z[int(z.shape[0] / 2):])  # ´edge_label_index[1]´
        return edge_recon_logits


import numpy as np
from typing import List, Literal, Optional, Tuple, Union

import torch
import torch.nn as nn
from torch import Tensor
from torch.nn.functional import normalize
from torch_geometric.data import Data
from torch_geometric.nn import GAE

from .utils import extract_subgraph, to_float_tensor
from ..nn.decoders import GATDecoder, GCNDecoder, CosineSimGraphDecoder
from .loss import compute_omics_recon_mse_loss, compute_omics_recon_mmd_loss, compute_edge_recon_loss, \
    compute_kl_reg_loss, compute_contrastive_instanceloss, compute_contrastive_clusterloss

# Based on VGAE class in PyTorch Geometric
class GNNModelVAE(GAE):
    def __init__(self, encoder, bottle_neck_neurons, hidden_dims, feature_dim,
                 num_heads, dropout, concat, n_domain, used_edge_weight, used_DSBN,
                 conv_type, gnn_layer=2, cluster_num=20, include_edge_recon_loss=True,
                 include_gene_expr_recon_loss=True, used_mmd=False
                 ):
        """
        Parameters
        ----------
        encoder : nn.Module
            The encoder module used in the variational graph autoencoder. 'GAT' or 'GCN'.
        bottle_neck_neurons : int
            Number of neurons in the bottleneck layer representing the latent dimension.
        hidden_dims : int
            Number of hidden dimensions for the encoder.
        feature_dim : int
            Number of feature dimensions in the input data.
        num_heads : int
            Number of attention heads used in the GAT encoder.
        dropout : float
            Dropout rate used in the encoder and decoder.
        concat : bool
            Whether to concatenate outputs of different attention heads.
        n_domain : int
            Number of domains for domain-specific batch normalization (DSBN).
        used_edge_weight : bool
            Whether to use edge weights in the graph convolution operation.
        used_DSBN : bool
            Whether to use domain-specific batch normalization (DSBN).
        conv_type : str
            Type of graph convolution to use, e.g., 'GAT', 'GATv2Conv', 'GCN'.
        gnn_layer : int, optional
            Number of layers in the GNN encoder. Default is 2.
        cluster_num : int, optional
            Number of clusters for the clustering layer. Default is 20.
        include_edge_recon_loss : bool, optional
            Whether to include edge reconstruction loss in the model. Default is True.
        include_gene_expr_recon_loss : bool, optional
            Whether to include gene expression reconstruction loss in the model. Default is True.
        used_mmd : bool, optional
            Whether to use MMD (Maximum Mean Discrepancy) loss for domain adaptation. Default is False.
        """
        super(GNNModelVAE, self).__init__(encoder)

        # model configurations
        self.encoder = encoder
        self.latent = bottle_neck_neurons
        self.hidden_dims = hidden_dims
        self.feature_dim = feature_dim
        self.num_heads = num_heads
        self.dropout = dropout
        self.concat = concat
        self.n_domain = n_domain
        self.used_edge_weight = used_edge_weight
        self.used_DSBN = used_DSBN
        self.include_edge_recon_loss = include_edge_recon_loss
        self.include_gene_expr_recon_loss = include_gene_expr_recon_loss
        self.used_mmd = used_mmd
        self.conv_type = conv_type
        self.cluster_num = cluster_num
        assert self.conv_type in ['GAT', 'GATv2Conv', 'GCN'], 'Convolution must be "GCN", "GAT" or "GATv2Conv".'
        self.gnn_layer = gnn_layer

        # 使用 Xavier 初始化权重
        self.eps_weight = nn.Parameter(nn.init.xavier_uniform_(torch.empty(self.latent, self.latent)))
        self.eps_bias = nn.Parameter(torch.zeros(self.latent))
        self.instance_projector = nn.Sequential(
            nn.Linear(self.latent, self.latent),
            nn.LayerNorm(self.latent, elementwise_affine=False),
            nn.ReLU(),
            nn.Linear(self.latent, self.feature_dim),
        )
        self.cluster_projector = nn.Sequential(
            nn.Linear(self.latent, self.latent),
            nn.LayerNorm(self.latent, elementwise_affine=False),
            nn.ReLU(),
            nn.Linear(self.latent, self.cluster_num),
            nn.Softmax(dim=1)
        )
        # Initialize graph decoder module
        self.graph_decoder = CosineSimGraphDecoder(
            dropout_rate=self.dropout)

        ## 重构表达谱
        if self.conv_type in ['GAT', 'GATv2Conv']:
            self.GAT_decoder = GATDecoder(
                in_channels=self.latent,
                hidden_dims=self.hidden_dims,
                out_channels=self.feature_dim,
                conv_type=self.conv_type,
                num_heads=self.num_heads,
                dropout=self.dropout,
                concat=self.concat,
                num_domains=self.n_domain,  # DSBN
                used_edge_weight=self.used_edge_weight,
                used_DSBN=self.used_DSBN
            )
        elif self.conv_type == 'GCN':
            self.GCN_decoder = GCNDecoder(
                in_channels=self.latent,
                hidden_dims=self.hidden_dims,
                out_channels=self.feature_dim,
                dropout=self.dropout,
                num_domains=self.n_domain,  # DSBN
                used_edge_weight=self.used_edge_weight,
                used_DSBN=self.used_DSBN
            )
        else:
            raise NotImplementedError("Unknown GNN-Operator.")

    def reparameterize(self, mu: Tensor, logstd: Tensor, eps=None) -> Tensor:
        """
        Applies the reparameterization trick to sample a latent vector from the latent distribution during training.

        Parameters
        ----------
        mu : torch.Tensor
            Mean of the latent distribution (shape: [batch_size, latent_dim]).
        logstd : torch.Tensor
            Logarithm of the standard deviation of the latent distribution (shape: [batch_size, latent_dim]).
        eps : torch.Tensor, optional
            Noise tensor used for sampling. If not provided, a standard normal distribution will be used (shape: [batch_size, latent_dim]).

        Returns
        -------
        torch.Tensor
            A reparameterized latent vector sampled from the distribution (shape: [batch_size, latent_dim]).
        """


        if self.training:
            if eps is not None:
                return mu + eps * torch.randn_like(logstd) * torch.exp(logstd)
            else:
                return mu + torch.randn_like(logstd) * torch.exp(logstd)
        else:
            return mu

    def forward(self, data_batch, decoder_type, augment_type):
        """
        Processes the input data through the encoder to obtain the latent representations and uses the decoder to reconstruct features or edges, depending on the task.

        Parameters
        ----------
        data_batch : Data
            A PyTorch Geometric Data object containing node features, edge information, and any other relevant data.
        decoder_type : str
            Specifies which type of decoder to use, either 'omics' for gene expression data or 'graph' for edge reconstruction tasks.
        augment_type : str
            Specifies the type of data augmentation used during encoding, e.g., 'svd' for singular value decomposition or 'dropout' for regular dropout.

        Returns
        -------
        dict
            A dictionary containing the following keys:
            - "recon_features" or "edge_recon_logits": Reconstructed features or edge logits, depending on the decoder type.
            - "z": The latent representation of the input data.
            - "mu": The mean of the latent distribution.
            - "logstd": The log standard deviation of the latent distribution.
            - "truth_x": Ground truth input features (for omics tasks).
            - "truth_y": Ground truth labels (for MMD or classification tasks).
        """
        # Get index of sampled nodes for current batch (neighbors of sampled
        # nodes are also part of the batch for message passing layers but
        # should be excluded in backpropagation)
        if decoder_type == "omics":
            # ´data_batch´ will be a node batch and first node_batch_size
            # elements are the sampled nodes, leading to a dim of ´batch_idx´ of
            # ´node_batch_size´
            # batch_idx = slice(None, data_batch.batch_size)

            all_mu1 = []
            all_mu2 = []
            for _ in range(self.gnn_layer):
                data_batch.x = to_float_tensor(data_batch.x)
                encoder_outputs = self.encoder(data_batch, decoder_type, augment_type)
                mu1 = encoder_outputs[0] #[batch_idx, :]
                mu2 = encoder_outputs[2] #[batch_idx, :]
                all_mu1.append(mu1)
                all_mu2.append(mu2)

            mean1 = torch.stack(all_mu1).mean(dim=0)  # sum
            logstd1 = torch.matmul(mean1, self.eps_weight) + self.eps_bias
            z1 = self.reparameterize(mean1, logstd1)

            mean2 = torch.stack(all_mu2).mean(dim=0)
            logstd2 = torch.matmul(mean2, self.eps_weight) + self.eps_bias
            z2 = self.reparameterize(mean2, logstd2)

            z_1 = normalize(self.instance_projector(z1), dim=1)
            z_2 = normalize(self.instance_projector(z2), dim=1)

            c_1 = self.cluster_projector(z1)
            c_2 = self.cluster_projector(z2)

            ## 重构表达矩阵
            output = {}
            # with torch.no_grad(): ## TODO
            # data_batch = extract_subgraph(data_batch, batch_idx)
            if self.conv_type in ['GAT', 'GATv2Conv']:
                recon_features = self.GAT_decoder(z1, data_batch)
            else:
                recon_features = self.GCN_decoder(z1, data_batch)
            output["truth_x"] = to_float_tensor(data_batch.x)
            output["truth_y"] = data_batch.y
            output["recon_features"] = to_float_tensor(recon_features)

            output["z"] = z1
            output["z_1"] = z_1
            output["z_2"] = z_2
            output["c_1"] = c_1
            output["c_2"] = c_2
            output["mu"] = mean1
            output["logstd"] = logstd1
            return output

        elif decoder_type == "graph":
            # ´data_batch´ will be an edge batch with sampled positive and
            # negative edges of size ´edge_batch_size´ respectively. Each edge
            # has a source and target node, leading to a dim of ´batch_idx´ of
            # 4 * ´edge_batch_size´
            batch_idx = torch.cat((data_batch.edge_label_index[0],
                                   data_batch.edge_label_index[1]), 0)

            all_mu1 = []
            for _ in range(self.gnn_layer):
                data_batch.x = to_float_tensor(data_batch.x)
                encoder_outputs = self.encoder(data_batch, decoder_type, augment_type)
                mu1 = encoder_outputs[0][batch_idx, :]
                all_mu1.append(mu1)

            mean1 = torch.stack(all_mu1).mean(dim=0)  # sum
            logstd1 = torch.matmul(mean1, self.eps_weight) + self.eps_bias
            z1 = self.reparameterize(mean1, logstd1)

            ## 重构表达矩阵
            output = {}
            # Store edge labels in output for loss computation
            output["edge_recon_labels"] = data_batch.edge_label

            # Use decoder to get the edge reconstruction logits
            output["edge_recon_logits"] = self.graph_decoder(z1)

            output["z"] = z1
            output["mu"] = mean1
            output["logstd"] = logstd1
            return output



    def loss(self,
             edge_model_output: dict,
             node_model_output: dict,
             lambda_edge_recon,
             lambda_gene_expr_recon,
             lambda_latent_contrastive_instanceloss,
             lambda_latent_contrastive_clusterloss,
             lambda_omics_recon_mmd_loss) -> dict:
        """
        Computes the total loss for the model by combining different loss components such as KL divergence, edge reconstruction loss, gene expression reconstruction loss, and contrastive losses.

        Parameters
        ----------
        edge_model_output : dict
            A dictionary containing outputs from the forward pass for edge reconstruction, including edge logits and latent variables.
        node_model_output : dict
            A dictionary containing outputs from the forward pass for node reconstruction, including gene expression reconstruction and latent variables.
        lambda_edge_recon : float
            A scaling factor to adjust the contribution of edge reconstruction loss.
        lambda_gene_expr_recon : float
            A scaling factor to adjust the contribution of gene expression reconstruction loss.
        lambda_latent_contrastive_instanceloss : float
            A scaling factor to adjust the contribution of instance-level contrastive loss between different latent representations.
        lambda_latent_contrastive_clusterloss : float
            A scaling factor to adjust the contribution of cluster-level contrastive loss between different clusters in the latent space.
        lambda_omics_recon_mmd_loss : float
            A scaling factor to adjust the contribution of Maximum Mean Discrepancy (MMD) loss in omics data reconstruction.

        Returns
        -------
        dict
            A dictionary containing individual loss terms and the total loss:
            - "kl_reg_loss": KL divergence loss between the latent distributions.
            - "edge_recon_loss": Binary cross-entropy loss for edge reconstruction (if applicable).
            - "gene_expr_recon_loss": Mean squared error loss for gene expression reconstruction (if applicable).
            - "lambda_latent_contrastive_instanceloss": Contrastive loss between instance-level latent vectors.
            - "lambda_latent_contrastive_clusterloss": Contrastive loss between clusters in the latent space.
            - "gene_expr_mmd_loss": MMD loss for omics data (if applicable).
            - "global_loss": Sum of all the individual losses used for model optimization.
            - "optim_loss": Sum of the losses used for backpropagation.
        """
        loss_dict = {}

        # 1. Compute Kullback-Leibler divergence loss for edge and node batch
        loss_dict["kl_reg_loss"] = compute_kl_reg_loss(
            mu=node_model_output["mu"],
            logstd=node_model_output["logstd"])
        loss_dict["kl_reg_loss"] += compute_kl_reg_loss(
            mu=edge_model_output["mu"],
            logstd=edge_model_output["logstd"])

        # 2. Compute edge reconstruction binary cross entropy loss for edge batch
        loss_dict["edge_recon_loss"] = (
                lambda_edge_recon * compute_edge_recon_loss(
            edge_recon_logits=edge_model_output["edge_recon_logits"],
            edge_recon_labels=edge_model_output["edge_recon_labels"]))

        # 3. Compute gene expression reconstruction with MSE loss for node batch
        loss_dict["gene_expr_recon_loss"] = (
                lambda_gene_expr_recon *
                compute_omics_recon_mse_loss(
                    recon_x=node_model_output["recon_features"],
                    x=node_model_output["truth_x"])) # * node_model_output['truth_x'].size(0)

        # 4. compute Contrastive instance losses
        loss_dict["lambda_latent_contrastive_instanceloss"] = compute_contrastive_instanceloss(node_model_output['z_1'],
                                                                                node_model_output['z_2'],
                                                                                lambda_latent_contrastive_instanceloss)
        # 5. compute Contrastive cluster losses
        loss_dict["lambda_latent_contrastive_clusterloss"] = compute_contrastive_clusterloss(node_model_output['c_1'],
                                                                                node_model_output['c_2'],
                                                                                self.cluster_num,
                                                                                lambda_latent_contrastive_clusterloss)

        # 6. compute MMD loss
        if self.used_mmd:
            cell_batch = node_model_output['truth_y']
            device = cell_batch.device
            cell_batch = cell_batch.detach().cpu()
            unique_groups, group_indices = np.unique(cell_batch, return_inverse=True)
            grouped_z_cell = {group: node_model_output['z'][group_indices == i] for i, group in enumerate(unique_groups)}
            group_labels = list(unique_groups)
            num_groups = len(group_labels)

            loss_dict["gene_expr_mmd_loss"] = torch.tensor(0, dtype=torch.float).to(device)
            for i in range(num_groups):
                for j in range(i + 1, num_groups):
                    z_i = grouped_z_cell[group_labels[i]]
                    z_j = grouped_z_cell[group_labels[j]]
                    mmd_loss_tmp = compute_omics_recon_mmd_loss(z_i, z_j) * 600 # * node_model_output['z'].size(0)
                    loss_dict["gene_expr_mmd_loss"] += mmd_loss_tmp * lambda_omics_recon_mmd_loss

        # Compute optimization loss used for backpropagation as well as global
        # loss used for early stopping of model training and best model saving
        loss_dict["global_loss"] = 0
        loss_dict["optim_loss"] = 0
        loss_dict["global_loss"] += loss_dict["kl_reg_loss"]
        loss_dict["optim_loss"] += loss_dict["kl_reg_loss"]
        if self.include_edge_recon_loss:
            loss_dict["global_loss"] += loss_dict["edge_recon_loss"]
            loss_dict["optim_loss"] += loss_dict["edge_recon_loss"]
        if self.include_gene_expr_recon_loss:
            loss_dict["global_loss"] += loss_dict["gene_expr_recon_loss"]
            loss_dict["optim_loss"] += loss_dict["gene_expr_recon_loss"]
            loss_dict["global_loss"] += loss_dict["lambda_latent_contrastive_instanceloss"]
            loss_dict["optim_loss"] += loss_dict["lambda_latent_contrastive_instanceloss"]
            loss_dict["global_loss"] += loss_dict["lambda_latent_contrastive_clusterloss"]
            loss_dict["optim_loss"] += loss_dict["lambda_latent_contrastive_clusterloss"]
            if self.used_mmd:
                loss_dict["global_loss"] += loss_dict["gene_expr_mmd_loss"]
                loss_dict["optim_loss"] += loss_dict["gene_expr_mmd_loss"]
        return loss_dict

    def get_latent_representation(
            self,
            node_batch: Data,
            augment_type: Literal['svd', 'dropout'] = 'svd',
            return_mu_std: bool = False
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """
        Encodes the input data into latent space, either returning the latent features (z) or
        the distribution parameters (mu and std) based on the input option.

        Parameters
        ----------
        node_batch : Data
            A PyTorch Geometric Data object containing features and graph structure for the node-level batch.
        augment_type : str, optional
            Specifies the type of augmentation used in the encoder, e.g., 'svd' (default) or 'dropout'.
        return_mu_std : bool, optional
            If True, the function returns the mean (mu) and standard deviation (std) of the latent distribution.
            Otherwise, it returns the reparameterized latent features (z). Default is False.

        Returns
        -------
        Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]
            - If `return_mu_std` is False, it returns the reparameterized latent features (z) (shape: [batch_size, latent_dim]).
            - If `return_mu_std` is True, it returns a tuple of the mean (mu) and standard deviation (std) of the latent distribution (each of shape: [batch_size, latent_dim]).
        """

        # Get latent distribution parameters
        encoder_outputs = self.encoder(node_batch,
                                       augment_type=augment_type,
                                       decoder_type='omics') # z_mean1, z_log_std1, z_mean2, z_log_std2
        mu = encoder_outputs[0][:node_batch.batch_size, :]
        logstd = encoder_outputs[1][:node_batch.batch_size, :]

        if return_mu_std:
            std = torch.exp(logstd)
            return mu, std
        else:
            z = self.reparameterize(mu, logstd)
            return z


import torch
from torch_geometric.data import Data

def extract_subgraph(data, node_indices):
    # 如果 node_indices 是 slice 类型，转换为具体的索引列表
    if isinstance(node_indices, slice):
        node_indices = torch.arange(data.num_nodes)[node_indices]

    # Check if node_indices is a tensor, if not convert it
    if not isinstance(node_indices, torch.Tensor):
        node_indices = torch.tensor(node_indices, dtype=torch.long)

    # 提取这些节点的特征
    sub_x = data.x[node_indices]
    if data.y is not None:
        sub_y = data.y[node_indices]
    else:
        sub_y = None

    # 提取包含这些节点的边
    edge_index = data.edge_index
    node_indices = node_indices.to(edge_index.device)
    mask = torch.isin(edge_index[0], node_indices) & torch.isin(edge_index[1], node_indices)
    sub_edge_index = edge_index[:, mask]

    # 将边索引调整为新的节点索引
    _, new_indices = node_indices.unique(return_inverse=True)
    remap = {old_idx.item(): new_idx for old_idx, new_idx in zip(node_indices, new_indices)}
    for i in range(sub_edge_index.shape[1]):
        sub_edge_index[0, i] = remap[sub_edge_index[0, i].item()]
        sub_edge_index[1, i] = remap[sub_edge_index[1, i].item()]

    # 提取对应的边特征
    sub_edge_attr = data.edge_attr[mask] if data.edge_attr is not None else None

    return Data(x=sub_x, edge_index=sub_edge_index, y=sub_y, edge_attr=sub_edge_attr)

def to_float_tensor(tensor):
    # 检查张量的数据类型是否为浮点类型
    if not torch.is_floating_point(tensor):
        tensor = tensor.float()
        # 如果不是浮点类型，则将其转换为浮点类型
        if tensor.dtype != torch.float32:
            tensor = tensor.to(dtype=torch.float32)
        return tensor
    # 如果已经是浮点类型，判断是否为32位，否则直接返回原张量
    if tensor.dtype != torch.float32:
        tensor = tensor.to(dtype=torch.float32)
    return tensor



import warnings
from types import SimpleNamespace
from typing import Literal, List, Optional, Tuple, Union

import torch
import numpy as np
from anndata import AnnData

from .._settings import settings
from .basemodelmixin import BaseModelMixin
from ..data.dataprocessors import prepare_data
from ..preprocessing.preprocess import DataProcess
from ..data.dataloaders import initialize_dataloaders
from .utils import weighted_knn_trainer, weighted_knn_transfer
from ..nn.encoders import GATEncoder, GCNEncoder
from ..modules.GNNModelVAE import GNNModelVAE
from ..trainer.trainer import GarfieldTrainer

class Garfield(torch.nn.Module, BaseModelMixin):
    """
    Garfield: Graph-based Contrastive Learning enable Fast Single-Cell Embedding
    """
    def __init__(self, gf_params):
        """"
        Garfield Model Parameters, which are included in gf_params.

        Parameters
        ----------
        adata_list : list
            List of AnnData objects containing data from multiple batches or samples.
        profile : str
            Specifies the data profile type (e.g., 'RNA', 'ATAC', 'ADT', 'multi-modal', 'spatial').
        data_type : str
            Type of the multi-omics dataset (e.g., Paired, UnPaired) for preprocessing.
        sub_data_type : list[str]
            List of data types for multi-modal datasets (e.g., ['rna', 'atac'] or ['rna', 'adt']).
        sample_col : str
            Column in the dataset that indicates batch or sample identifiers (default: 'batch').
        weight : float or None
            Weighting factor that determines the contribution of different modalities or types of graphs in multi-omics or spatial data.
            - For non-spatial single-cell multi-omics data (e.g., RNA + ATAC),
            `weight` specifies the contribution of the graph constructed from RNA data.
            The remaining (1 - weight) represents the contribution from the other modality.
            - For spatial single-modality data,
            `weight` refers to the contribution of the graph constructed from the physical spatial information,
            while (1 - weight) reflects the contribution from the molecular graph.
        graph_const_method : str
            Method for constructing the graph (e.g., 'mu_std', 'Radius', 'KNN', 'Squidpy').
        genome : str
            Reference genome to use during preprocessing (e.g., 'mm10', 'mm9', 'hg38', 'hg19').
        use_gene_weight : bool
            Whether to apply gene weights in the preprocessing step.
        use_top_pcs : bool
            Whether to use the top principal components during gene score preprocessing step.
        used_hvg : bool
            Whether to use highly variable genes (HVGs) for analysis.
        min_features : int
            Minimum number of features required for a cell to be included in the dataset.
        min_cells : int
            Minimum number of cells required for a feature to be retained in the dataset.
        keep_mt : bool
            Whether to retain mitochondrial genes in the analysis.
        target_sum : float
            Target sum used for normalization (e.g., 1e4 for counts per cell).
        rna_n_top_features : int
            Number of top features to retain for RNA datasets (e.g., 3000).
        atac_n_top_features : int
            Number of top features to retain for ATAC datasets (e.g., 10000).
        n_components : int
            Number of components to use for dimensionality reduction (e.g., PCA).
        n_neighbors : int
            Number of neighbors to use in graph-based algorithms (e.g., KNN).
        metric : str
            Distance metric used during graph construction (e.g., 'correlation', 'euclidean').
        svd_solver : str
            Solver for singular value decomposition (SVD), such as 'arpack' or 'randomized'.
        adj_key : str
            Key in the AnnData object that holds the adjacency matrix.
        edge_val_ratio : float
            Ratio of edges to use for validation in edge-level tasks.
        edge_test_ratio : float
            Ratio of edges to use for testing in edge-level tasks.
        node_val_ratio : float
            Ratio of nodes to use for validation in node-level tasks.
        node_test_ratio : float
            Ratio of nodes to use for testing in node-level tasks.
        augment_type : str
            Type of augmentation to use (e.g., 'dropout', 'svd').
        svd_q : int
            Rank for the low-rank SVD approximation.
        use_FCencoder : bool
            Whether to use a fully connected encoder before the graph layers.
        hidden_dims : list[int]
            List of hidden layer dimensions for the encoder.
        bottle_neck_neurons : int
            Number of neurons in the bottleneck (latent) layer.
        num_heads : int
            Number of attention heads for each graph attention layer.
        dropout : float
            Dropout rate applied during training.
        concat : bool
            Whether to concatenate attention heads or not.
        drop_feature_rate : float
            Dropout rate applied to node features.
        drop_edge_rate : float
            Dropout rate applied to edges during augmentation.
        used_edge_weight : bool
            Whether to use edge weights in the graph layers.
        used_DSBN : bool
            Whether to use domain-specific batch normalization.
        conv_type : str
            Type of graph convolution to use ('GATv2Conv', 'GAT', 'GCN').
        gnn_layer : int
            Number of times the encoder is repeated in the forward pass, not the number of GNN layers.
        cluster_num : int
            Number of clusters for latent feature clustering.
        num_neighbors : int
            Number of neighbors to sample for graph-based data loaders.
        loaders_n_hops : int
            Number of hops for neighbors during graph construction.
        edge_batch_size : int
            Batch size for edge-level tasks.
        node_batch_size : int
            Batch size for node-level tasks.
        include_edge_recon_loss : bool
            Whether to include edge reconstruction loss in the training objective.
        include_gene_expr_recon_loss : bool
            Whether to include gene expression reconstruction loss in the training objective.
        used_mmd : bool
            Whether to use maximum mean discrepancy (MMD) for domain adaptation.
        lambda_latent_contrastive_instanceloss : float
            Weight for the instance-level contrastive loss.
        lambda_latent_contrastive_clusterloss : float
            Weight for the cluster-level contrastive loss.
        lambda_gene_expr_recon : float
            Weight for the gene expression reconstruction loss.
        lambda_edge_recon : float
            Weight for the edge reconstruction loss.
        lambda_omics_recon_mmd_loss : float
            Weight for the MMD loss in omics reconstruction tasks.
        n_epochs : int
            Number of training epochs.
        n_epochs_no_edge_recon : int
            Number of epochs without edge reconstruction loss.
        learning_rate : float
            Learning rate for the optimizer.
        weight_decay : float
            Weight decay (L2 regularization) for the optimizer.
        gradient_clipping : float
            Maximum norm for gradient clipping.
        latent_key : str
            Key for storing latent features in the AnnData object.
        reload_best_model : bool
            Whether to reload the best model after training.
        use_early_stopping : bool
            Whether to use early stopping during training.
        early_stopping_kwargs : dict
            Arguments for configuring early stopping (e.g., patience, delta).
        monitor : bool
            Whether to print training progress.
        seed : int
            Random seed for reproducibility.
        verbose : bool
            Whether to display detailed logs during training.
        """
        super(Garfield, self).__init__()
        if gf_params is None:
            gf_params = settings.gf_params.copy()
        else:
            assert isinstance(gf_params, dict), "`gf_params` must be dict"

        self.args = SimpleNamespace(**gf_params)
        # data preprocessing parameters
        self.adata_list_ = self.args.adata_list
        self.profile_ = self.args.profile
        self.data_type_ = self.args.data_type
        self.sub_data_type_ = self.args.sub_data_type
        self.sample_col_ = self.args.sample_col
        self.weight_ = self.args.weight
        self.graph_const_method_ = self.args.graph_const_method
        self.genome_ = self.args.genome
        self.use_gene_weight_ = self.args.use_gene_weight
        self.use_top_pcs_ = self.args.use_top_pcs
        self.used_hvg_ = self.args.used_hvg
        self.min_features_ = self.args.min_features
        self.min_cells_ = self.args.min_cells
        self.keep_mt_ = self.args.keep_mt
        self.target_sum_ = self.args.target_sum
        self.rna_n_top_features_ = self.args.rna_n_top_features
        self.atac_n_top_features_ = self.args.atac_n_top_features
        self.n_components_ = self.args.n_components
        self.n_neighbors_ = self.args.n_neighbors
        self.metric_ = self.args.metric
        self.svd_solver_ = self.args.svd_solver
        # datasets
        self.used_pca_feat_ = self.args.used_pca_feat
        self.adj_key_ = self.args.adj_key
        # data split parameters
        self.edge_val_ratio_ = self.args.edge_val_ratio
        self.edge_test_ratio_ = self.args.edge_test_ratio
        self.node_val_ratio_ = self.args.node_val_ratio
        self.node_test_ratio_ = self.args.node_test_ratio
        # model parameters
        self.augment_type_ = self.args.augment_type
        self.svd_q_ = self.args.svd_q  # if augment_type == 'svd'
        self.use_FCencoder_ = self.args.use_FCencoder
        self.hidden_dims_ = self.args.hidden_dims
        self.bottle_neck_neurons_ = self.args.bottle_neck_neurons
        self.num_heads_ = self.args.num_heads
        self.dropout_ = self.args.dropout
        self.concat_ = self.args.concat
        self.drop_feature_rate_ = self.args.drop_feature_rate
        self.drop_edge_rate_ = self.args.drop_edge_rate
        self.used_edge_weight_ = self.args.used_edge_weight
        self.used_DSBN_ = self.args.used_DSBN
        # self.n_domain_ = self.args.num_classes
        self.conv_type_ = self.args.conv_type
        self.gnn_layer_ = self.args.gnn_layer
        self.cluster_num_ = self.args.cluster_num
        # data loader parameters
        self.num_neighbors_ = self.args.num_neighbors
        self.loaders_n_hops_ = self.args.loaders_n_hops
        self.edge_batch_size_ = self.args.edge_batch_size
        self.node_batch_size_ = self.args.node_batch_size
        # loss parameters
        self.include_edge_recon_loss_ = self.args.include_edge_recon_loss
        self.include_gene_expr_recon_loss_ = self.args.include_gene_expr_recon_loss
        self.used_mmd_ = self.args.used_mmd
        self.lambda_latent_contrastive_instanceloss_ = self.args.lambda_latent_contrastive_instanceloss
        self.lambda_latent_contrastive_clusterloss_ = self.args.lambda_latent_contrastive_clusterloss
        self.lambda_gene_expr_recon_ = self.args.lambda_gene_expr_recon
        self.lambda_edge_recon_ = self.args.lambda_edge_recon
        self.lambda_omics_recon_mmd_loss_ = self.args.lambda_omics_recon_mmd_loss
        # train parameters
        self.n_epochs_ = self.args.n_epochs
        self.n_epochs_no_edge_recon_ = self.args.n_epochs_no_edge_recon
        self.learning_rate_ = self.args.learning_rate
        self.weight_decay_ = self.args.weight_decay
        self.gradient_clipping_ = self.args.gradient_clipping
        # other parameters
        self.latent_key_ = self.args.latent_key
        self.reload_best_model_ = self.args.reload_best_model
        self.use_early_stopping_ = self.args.use_early_stopping
        self.early_stopping_kwargs_ = self.args.early_stopping_kwargs
        self.monitor_ = self.args.monitor
        self.seed_ = self.args.seed
        self.verbose_ = self.args.verbose

        # Set seed for reproducibility
        np.random.seed(self.seed_)
        if torch.cuda.is_available():
            torch.cuda.manual_seed(self.seed_)
            torch.manual_seed(self.seed_)
        else:
            torch.manual_seed(self.seed_)

        # Data load and preprocessing
        print("--- DATA LOADING AND PREPROCESSING ---")
        self.adata = DataProcess(
            adata_list=self.adata_list_,
            profile=self.profile_,
            data_type=self.data_type_,
            sub_data_type=self.sub_data_type_,
            sample_col=self.sample_col_,
            genome=self.genome_,
            weight=self.weight_,
            graph_const_method=self.graph_const_method_,
            use_gene_weigt=self.use_gene_weight_,
            use_top_pcs=self.use_top_pcs_,
            used_hvg=self.used_hvg_,
            min_features=self.min_features_,
            min_cells=self.min_cells_,
            keep_mt=self.keep_mt_,
            target_sum=self.target_sum_,
            rna_n_top_features=self.rna_n_top_features_,
            atac_n_top_features=self.atac_n_top_features_,
            n_components=self.n_components_,
            n_neighbors=self.n_neighbors_,
            metric=self.metric_,
            svd_solver=self.svd_solver_
        )
        # set up model
        if not self.used_pca_feat_:
            self.num_features_ = self.adata.n_vars
        else:
            self.num_features_ = self.adata.obsm['feat'].shape[1]

        if self.sample_col_ is not None:
            self.n_domain_ = len(self.adata.obs[self.sample_col_].unique())
        else:
            self.n_domain_ = None
        self.setup_layers()

    def setup_layers(self):
        """
        Creating the layers.
        """
        ## 选择 encoder
        ## 设定参数
        if self.conv_type_ in ['GAT', 'GATv2Conv']:
            encoder = GATEncoder(
                in_channels=self.num_features_,
                hidden_dims=self.hidden_dims_,
                latent_dim=self.bottle_neck_neurons_,
                conv_type=self.conv_type_,
                use_FCencoder=self.use_FCencoder_,
                num_heads=self.num_heads_,
                dropout=self.dropout_,
                concat=self.concat_,
                num_domains=1,  # batch normalization
                drop_feature_rate=self.drop_feature_rate_,
                drop_edge_rate=self.drop_edge_rate_,
                used_edge_weight=self.used_edge_weight_,
                svd_q=self.svd_q_,
                used_DSBN=self.used_DSBN_
            )
        else:
            encoder = GCNEncoder(
                in_channels=self.num_features_,
                hidden_dims=self.hidden_dims_,
                latent_dim=self.bottle_neck_neurons_,
                use_FCencoder=self.use_FCencoder_,
                dropout=self.dropout_,
                num_domains=1,  # batch normalization
                drop_feature_rate=self.drop_feature_rate_,
                drop_edge_rate=self.drop_edge_rate_,
                used_edge_weight=self.used_edge_weight_,
                svd_q=self.svd_q_,
                used_DSBN=self.used_DSBN_
            )

        ## GCNModelVAE
        self.model = GNNModelVAE(
            encoder=encoder,
            bottle_neck_neurons=self.bottle_neck_neurons_,
            hidden_dims=self.hidden_dims_,
            feature_dim=self.num_features_,
            num_heads=self.num_heads_,
            dropout=self.dropout_,
            concat=self.concat_,
            n_domain=self.n_domain_,
            used_edge_weight=self.used_edge_weight_,
            used_DSBN=self.used_DSBN_,
            conv_type=self.conv_type_,
            gnn_layer=self.gnn_layer_,
            cluster_num=self.cluster_num_,
            include_edge_recon_loss=self.include_edge_recon_loss_,
            include_gene_expr_recon_loss=self.include_gene_expr_recon_loss_,
            used_mmd=self.used_mmd_
        )
        self.is_trained_ = False

    def train(self, **trainer_kwargs):
        self.trainer = GarfieldTrainer(
            adata=self.adata,
            model=self.model,
            label_name=self.sample_col_,
            used_pca_feat=self.used_pca_feat_,
            adj_key=self.adj_key_,
            # data split
            edge_val_ratio=self.edge_val_ratio_,
            edge_test_ratio=self.edge_test_ratio_,
            node_val_ratio=self.node_val_ratio_,
            node_test_ratio=self.node_test_ratio_,
            # data process
            augment_type=self.augment_type_,
            # data loader
            num_neighbors=self.num_neighbors_,
            loaders_n_hops=self.loaders_n_hops_,
            edge_batch_size=self.edge_batch_size_,
            node_batch_size=self.node_batch_size_,
            # other parameters
            reload_best_model=self.reload_best_model_,
            use_early_stopping=self.use_early_stopping_,
            early_stopping_kwargs=self.early_stopping_kwargs_,
            monitor=self.monitor_,
            verbose=self.verbose_,
            seed=self.seed_,
            **trainer_kwargs)

        self.trainer.train(
            n_epochs=self.n_epochs_,
            n_epochs_no_edge_recon=self.lambda_edge_recon_,  # : int=0
            learning_rate=self.learning_rate_,
            weight_decay=self.weight_decay_,
            gradient_clipping=self.gradient_clipping_,
            lambda_edge_recon=self.lambda_edge_recon_,
            lambda_gene_expr_recon=self.lambda_gene_expr_recon_,
            lambda_latent_contrastive_instanceloss=self.lambda_latent_contrastive_instanceloss_,
            lambda_latent_contrastive_clusterloss=self.lambda_latent_contrastive_clusterloss_,
            lambda_omics_recon_mmd_loss=self.lambda_omics_recon_mmd_loss_
        )

        self.node_batch_size_ = self.trainer.node_batch_size_

        self.is_trained_ = True
        self.model.eval()

        self.adata.obsm[self.latent_key_], _ = self.get_latent_representation(
            adata=self.adata,
            adj_key=self.adj_key_,
            return_mu_std=True,
            node_batch_size=self.node_batch_size_
        )

    # embedding
    def get_latent_representation(
            self,
            adata: Optional[AnnData] = None,
            adj_key: str = "spatial_connectivities",
            return_mu_std: bool = False,
            node_batch_size: int = 64,
            dtype: type = np.float64,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """
        Get the latent representation / gene program scores from a trained model.

        Parameters
        ----------
        adata:
            AnnData object to get the latent representation for. If ´None´, uses
            the adata object stored in the model instance.
        adj_key:
            Key under which the sparse adjacency matrix is stored in
            ´adata.obsp´.
        return_mu_std:
            If `True`, return ´mu´ and ´std´ instead of latent features ´z´.
        node_batch_size:
            Batch size used during data loading.
        dtype:
            Precision to store the latent representations.

        Returns
        ----------
        z:
            Latent space features (dim: n_obs x n_active_gps or n_obs x n_gps).
        mu:
            Expected values of the latent posterior (dim: n_obs x n_active_gps
            or n_obs x n_gps).
        std:
            Standard deviations of the latent posterior (dim: n_obs x
            n_active_gps or n_obs x n_gps).
        """
        self._check_if_trained(warn=False)

        device = next(self.model.parameters()).device

        if adata is None:
            adata = self.adata

        # Create single dataloader containing entire dataset
        data_dict = prepare_data(
            adata=adata,
            adj_key=adj_key,
            used_pca_feat=self.used_pca_feat_,
            edge_val_ratio=0.,
            edge_test_ratio=0.,
            node_val_ratio=0.,
            node_test_ratio=0.)
        node_masked_data = data_dict["node_masked_data"]
        loader_dict = initialize_dataloaders(
            node_masked_data=node_masked_data,
            edge_train_data=None,
            edge_val_data=None,
            edge_batch_size=None,
            node_batch_size=node_batch_size,
            shuffle=False)
        node_loader = loader_dict["node_train_loader"]

        # Initialize latent vectors
        if return_mu_std:
            mu = np.empty(shape=(adata.shape[0], self.bottle_neck_neurons_), dtype=dtype)
            std = np.empty(shape=(adata.shape[0], self.bottle_neck_neurons_), dtype=dtype)
        else:
            z = np.empty(shape=(adata.shape[0], self.bottle_neck_neurons_), dtype=dtype)

        # Get latent representation for each batch of the dataloader and put it
        # into latent vectors
        for i, node_batch in enumerate(node_loader):
            n_obs_before_batch = i * node_batch_size
            n_obs_after_batch = n_obs_before_batch + node_batch.batch_size
            node_batch = node_batch.to(device)
            if return_mu_std:
                mu_batch, std_batch = self.model.get_latent_representation(
                    node_batch=node_batch,
                    augment_type=self.augment_type_,
                    return_mu_std=True)
                mu[n_obs_before_batch:n_obs_after_batch, :] = (
                    mu_batch.detach().cpu().numpy())
                std[n_obs_before_batch:n_obs_after_batch, :] = (
                    std_batch.detach().cpu().numpy())
            else:
                z_batch = self.model.get_latent_representation(
                    node_batch=node_batch,
                    augment_type=self.augment_type_,
                    return_mu_std=False)
                z[n_obs_before_batch:n_obs_after_batch, :] = (
                    z_batch.detach().cpu().numpy())
        if return_mu_std:
            return mu, std
        else:
            return z

    # Loss curve
    def plot_loss_curves(self,
                         title="Losses Curve"
                         ):
        return self.trainer.plot_loss_curves(title=title)

    @classmethod
    def _get_init_params_from_dict(cls, dct):
        init_params = {
            # Preprocessing options
            'adata_list': dct['adata_list_'],
            'profile': dct['profile_'],
            'data_type': dct['data_type_'],
            'sub_data_type': dct['sub_data_type_'],
            'sample_col': dct['sample_col_'],
            'weight': dct['weight_'],
            'graph_const_method': dct['graph_const_method_'],
            'genome': dct['genome_'],
            'use_gene_weight': dct['use_gene_weight_'],
            'use_top_pcs': dct['use_top_pcs_'],
            'used_hvg': dct['used_hvg_'],
            'min_features': dct['min_features_'],
            'min_cells': dct['min_cells_'],
            'keep_mt': dct['keep_mt_'],
            'target_sum': dct['target_sum_'],
            'rna_n_top_features': dct['rna_n_top_features_'],
            'atac_n_top_features': dct['atac_n_top_features_'],
            'n_components': dct['n_components_'],
            'n_neighbors': dct['n_neighbors_'],
            'metric': dct['metric_'],
            'svd_solver': dct['svd_solver_'],
            'adj_key': dct['adj_key_'],
            # data split parameters
            'edge_val_ratio': dct['edge_val_ratio_'],
            'edge_test_ratio': dct['edge_test_ratio_'],
            'node_val_ratio': dct['node_val_ratio_'],
            'node_test_ratio': dct['node_test_ratio_'],
            # model parameters
            'augment_type': dct['augment_type_'],
            'svd_q': dct['svd_q_'],
            'use_FCencoder': dct['use_FCencoder_'],
            'gnn_layer': dct['gnn_layer_'],
            'conv_type': dct['conv_type_'],
            'hidden_dims': dct['hidden_dims_'],
            'bottle_neck_neurons': dct['bottle_neck_neurons_'],
            'cluster_num': dct['cluster_num_'],
            'num_heads': dct['num_heads_'],
            'dropout': dct['dropout_'],
            'concat': dct['concat_'],
            'drop_feature_rate': dct['drop_feature_rate_'],
            'drop_edge_rate': dct['drop_edge_rate_'],
            'used_edge_weight': dct['used_edge_weight_'],
            'used_DSBN': dct['used_DSBN_'],
            'used_mmd': dct['used_mmd_'],
            # data loader parameters
            'num_neighbors': dct['num_neighbors_'],
            'loaders_n_hops': dct['loaders_n_hops_'],
            'edge_batch_size': dct['edge_batch_size_'],
            'node_batch_size': dct['node_batch_size_'],
            # loss parameters
            'include_edge_recon_loss': dct['include_edge_recon_loss_'],
            'include_gene_expr_recon_loss': dct['include_gene_expr_recon_loss_'],
            'lambda_latent_contrastive_instanceloss': dct['lambda_latent_contrastive_instanceloss_'],
            'lambda_latent_contrastive_clusterloss': dct['lambda_latent_contrastive_clusterloss_'],
            'lambda_gene_expr_recon': dct['lambda_gene_expr_recon_'],
            'lambda_edge_recon': dct['lambda_edge_recon_'],
            'lambda_omics_recon_mmd_loss': dct['lambda_omics_recon_mmd_loss_'],
            # train parameters
            'n_epochs': dct['n_epochs_'],
            'n_epochs_no_edge_recon': dct['n_epochs_no_edge_recon_'],
            'learning_rate': dct['learning_rate_'],
            'weight_decay': dct['weight_decay_'],
            'gradient_clipping': dct['gradient_clipping_'],
            # other parameters
            'latent_key': dct['latent_key_'],
            'reload_best_model': dct['reload_best_model_'],
            'use_early_stopping': dct['use_early_stopping_'],
            'early_stopping_kwargs': dct['early_stopping_kwargs_'],
            'monitor': dct['monitor_'],
            'seed': dct['seed_'],
            'verbose': dct['verbose_'],
        }

        return init_params

    def label_transfer(self,
                       ref_adata,
                       ref_adata_emb,
                       query_adata,
                       query_adata_emb,
                       ref_adata_obs,
                       label_keys,
                       n_neighbors=50,
                       threshold=1,
                       pred_unknown=False,
                       mode="package"):
        knn_transformer = weighted_knn_trainer(
            train_adata=ref_adata,
            train_adata_emb=ref_adata_emb,  # location of our joint embedding
            n_neighbors=n_neighbors,
        )

        labels, uncert = weighted_knn_transfer(
            query_adata=query_adata,
            query_adata_emb=query_adata_emb,  # location of our embedding, query_adata.X in this case
            label_keys=label_keys,  # (start of) obs column name(s) for which to transfer labels
            knn_model=knn_transformer,
            ref_adata_obs=ref_adata_obs,
            threshold=threshold,
            pred_unknown=pred_unknown,
            mode=mode
        )
        # 定义列名的映射
        cols = ref_adata_obs.columns[ref_adata_obs.columns.str.startswith(label_keys)]
        if pred_unknown:
            rename_mapping_labels = {col: f"transferred_{col}_filtered" for col in cols}
        else:
            rename_mapping_labels = {col: f"transferred_{col}_unfiltered" for col in cols}

        # 定义 uncertainty 映射
        rename_mapping_uncert = {col: f"transferred_{col}_uncert" for col in cols}

        # 重命名列并加入到 'query_adata.obs'
        query_adata.obs = query_adata.obs.join(
            labels.rename(columns=rename_mapping_labels)
        )

        # 重命名列并加入到 'query_adata.obs'
        query_adata.obs = query_adata.obs.join(
            uncert.rename(columns=rename_mapping_uncert)
        )
        ## 去除 query_adata obs 中 NA 的列
        query_adata.obs = query_adata.obs.dropna(axis=1, how='all')

        return query_adata






import warnings
import io
import pybedtools
import pkgutil
import numpy as np
import pandas as pd
import anndata as ad
from anndata import AnnData
import scipy.sparse
from scipy.sparse import (
    csr_matrix,
    coo_matrix,
)
from collections import defaultdict, Counter
from pandas.api.types import is_numeric_dtype
from sklearn.feature_extraction.text import TfidfTransformer
from scipy.sparse.linalg import svds
from sklearn.cross_decomposition import CCA
from sklearn.utils.extmath import randomized_svd

import logging
logger = logging.getLogger(__name__)


def reindex(adata, genes):
    """
    Reindex AnnData with gene list

    Parameters
    ----------
    adata
        AnnData
    genes
        gene list for indexing

    Return
    ------
    AnnData
    """
    # Warning for gene percentage
    user_var_names = adata.var_names
    user_var_names = user_var_names.astype(str)
    try:
        percentage = (len(user_var_names.intersection(genes)) / len(user_var_names)) * 100
        percentage = round(percentage, 4)
        if percentage != 100:
            logger.warning(f"WARNING: Query shares {percentage}% of its genes with the reference."
                           "This may lead to inaccuracy in the results.")
    except Exception:
        logger.warning("WARNING: Something is wrong with the reference genes.")

    # Get genes in reference that are not in query
    ref_genes_not_in_query = []
    for name in genes:
        if name not in user_var_names:
            ref_genes_not_in_query.append(name)

    if len(ref_genes_not_in_query) > 0:
        print("Query data is missing expression data of ",
              len(ref_genes_not_in_query),
              " genes which were contained in the reference dataset.")

    idx = [i for i, g in enumerate(genes) if g in adata.var_names]
    print('There are {} gene in selected genes'.format(len(idx)))
    if len(idx) == len(genes):
        adata = adata[:, genes]
    else:
        new_X = scipy.sparse.lil_matrix((adata.shape[0], len(genes)))
        new_X[:, idx] = adata[:, genes[idx]].X
        adata = AnnData(new_X.tocsr(), obs=adata.obs, var={'var_names': genes})
    return adata


def get_centroids(sub_data, labels):
    """
    Compute the centroids (cluster mean) of adata.

    Parameters
    ----------
    sub_data: scanpy object of each sample
    labels: np.array of shape (n_samples,)
        Cluster labels of each sample, coded from 0, ..., num_clusters-1

    Returns
    -------
    centroids: np.array of shape (n_centroids, n_features)
        Matrix of cluster centroids, the i-th column is the centroid of the i-th cluster
    """
    arr = sub_data.X
    arr_raw = sub_data.layers["counts"]
    meta = sub_data.obs
    cluster_label_to_indices = defaultdict(list)
    for i, l in enumerate(labels):
        cluster_label_to_indices[l].append(i)

    unique_labels = sorted(cluster_label_to_indices.keys())
    if not all(i == l for i, l in enumerate(unique_labels)):
        raise ValueError('labels must be coded in integers from 0, ..., n_clusters-1.')

    ## 对于每个簇，计算其均值，返回标准化的矩阵
    centroids = np.empty((len(unique_labels), arr.shape[1]))
    for curr_label, indices in cluster_label_to_indices.items():
        centroids[curr_label, :] = arr[indices, :].mean(axis=0)

    ## 对于每个簇，计算其均值，返回raw counts的矩阵
    counts = np.empty((len(unique_labels), arr.shape[1]))
    for curr_label, indices in cluster_label_to_indices.items():
        counts[curr_label, :] = arr_raw[indices, :].mean(axis=0)

    meta_info = np.empty((len(unique_labels), meta.shape[1]),
                         dtype=object)  # dtype=object 的意思是在 NumPy 数组中使用 Python 对象作为数据类型
    for curr_label, indices in cluster_label_to_indices.items():
        for col_idx in range(meta.shape[1]):
            if is_numeric_dtype(meta.iloc[indices, col_idx].dtype):
                # 如果是数值型数据
                meta_info[curr_label, col_idx] = meta.iloc[indices, col_idx].mean(axis=0)
            else:
                # 如果是字符型或其他类型 则通过投票找到最可能的label
                true_labels = meta.iloc[:, col_idx]  # .tolist()
                res = summarize_clustering(labels, curr_label, true_labels)
                meta_info[curr_label, col_idx] = res

    meta_info = pd.DataFrame(meta_info, columns=sub_data.obs.columns)

    for i, column in enumerate(meta_info.columns):
        if type(meta_info.iloc[:, i][0]) == list:
            meta_info.iloc[:, i] = sum(meta_info.iloc[:, i], [])
        else:
            meta_info.iloc[:, i] = meta_info.iloc[:, i]

    # use scanpy functions to do the adata_sub construction
    adata_sub = ad.AnnData(centroids, obs=meta_info, dtype=np.float32)
    adata_raw = ad.AnnData(counts, dtype=np.float32)
    adata_sub.layers["counts"] = adata_raw.X.copy()
    adata_sub.var_names = sub_data.var_names
    del adata_raw
    return adata_sub

## summarize_clustering
def summarize_clustering(clustering, curr_label, true_labels):
    """
    Compute the majority cell type for each cluster.

    Parameters
    ----------
    clustering: np.array of shape (n_samples,)
        Clustering labels, coded from 0, 1, ..., n_clusters.
    true_labels: np.array of shape (n_samples,)
        Groundtruth labels.

    Returns
    -------
    np.array of shape (n_clusters,)
        The majority voting results.
    """
    res = []
    cluster_label_to_indices = defaultdict(list)
    for i, l in enumerate(clustering):
        cluster_label_to_indices[l].append(i)

    # 使用字典推导式提取curr_label的子字典
    #     if len(curr_label) > 1:
    #         sub_dict = {key: cluster_label_to_indices[key] for key in curr_label}
    #     else:
    sub_dict = {curr_label: cluster_label_to_indices[curr_label]}

    for key in sub_dict.keys():
        curr_indices = sub_dict[key]  # sub_dict.get(key)
        curr_true_labels = true_labels.iloc[curr_indices]
        # majority voting
        # randomly break the ties
        curr_true_labels = np.random.permutation(curr_true_labels)
        counter = Counter(curr_true_labels)
        most_common_element, _ = counter.most_common(1)[0]
        res.append(most_common_element)

    return res


def cdist_correlation(arr1, arr2):
    """Calculate pair-wise 1 - Pearson correlation between X and Y.

    Parameters
    ----------
    arr1: np.array of shape (n_samples1, n_features)
        First dataset.
    arr2: np.array of shape (n_samples2, n_features)
        Second dataset.

    Returns
    -------
    array-like of shape (n_samples1, n_samples2)
        The (i, j)-th entry is 1 - Pearson correlation between i-th row of arr1 and j-th row of arr2.
    """
    n, p = arr1.shape
    m, p2 = arr2.shape
    assert p2 == p

    arr1 = (arr1.T - np.mean(arr1, axis=1)).T
    arr2 = (arr2.T - np.mean(arr2, axis=1)).T

    arr1 = (arr1.T / np.sqrt(1e-6 + np.sum(arr1 ** 2, axis=1))).T
    arr2 = (arr2.T / np.sqrt(1e-6 + np.sum(arr2 ** 2, axis=1))).T

    return 1 - arr1 @ arr2.T

def convert_to_numpy(arr):
    if isinstance(arr, csr_matrix):
        return arr.toarray()
    elif isinstance(arr, np.ndarray):
        return arr
    else:
        raise TypeError("Unsupported data type.")

def drop_zero_variability_columns(arr_lst: list, tol=1e-8):
    """
    Drop columns for which its standard deviation is zero in any one of the arrays in arr_list.

    Parameters
    ----------
    arr_lst: list of np.array
        List of arrays
    tol: float, default=1e-8
        Any number less than tol is considered as zero

    Returns
    -------
    List of np.array where no column has zero standard deviation
    """
    bad_columns = set()
    for arr in arr_lst:
        curr_std = np.std(arr, axis=0)
        for col in np.nonzero(np.abs(curr_std) < tol)[0]:
            bad_columns.add(col)
    good_columns = [i for i in range(arr_lst[0].shape[1]) if i not in bad_columns]
    return [arr[:, good_columns] for arr in arr_lst]


def robust_svd(arr, n_components, randomized=False, n_runs=1):
    """
    Do deterministic or randomized SVD on arr.

    Parameters
    ----------
    arr: np.array
        The array to do SVD on
    n_components: int
        Number of SVD components
    randomized: bool, default=False
        Whether to run randomized SVD
    n_runs: int, default=1
        Run multiple times and take the realization with the lowest Frobenious reconstruction error

    Returns
    -------
    u, s, vh: np.array
        u @ np.diag(s) @ vh is the reconstruction of the original arr
    """
    if randomized:
        best_err = float('inf')
        u, s, vh = None, None, None
        for _ in range(n_runs):
            curr_u, curr_s, curr_vh = randomized_svd(arr, n_components=n_components, random_state=None)
            curr_err = np.sum((arr - curr_u @ np.diag(curr_s) @ curr_vh) ** 2)
            if curr_err < best_err:
                best_err = curr_err
                u, s, vh = curr_u, curr_s, curr_vh
        assert u is not None and s is not None and vh is not None
    else:
        if n_runs > 1:
            warnings.warn("Doing deterministic SVD, n_runs reset to one.")
        u, s, vh = svds(arr*1.0, k=n_components) # svds can not handle integer values
    return u, s, vh

def svd_denoise(arr, n_components=20, randomized=False, n_runs=1):
    """
    Compute best rank-n_components approximation of arr by SVD.

    Parameters
    ----------
    arr: np.array of shape (n_samples, n_features)
        Data matrix
    n_components: int, default=20
        Number of components to keep
    randomized: bool, default=False
        Whether to use randomized SVD
    n_runs: int, default=1
        Run multiple times and take the realization with the lowest Frobenious reconstruction error

    Returns
    -------
    arr: array_like of shape (n_samples, n_features)
        Rank-n_comopnents approximation of the input arr.
    """
    if n_components is None:
        return arr
    u, s, vh = robust_svd(arr, n_components=n_components, randomized=randomized, n_runs=n_runs)
    return u @ np.diag(s) @ vh

def svd_embedding(arr, n_components=20, randomized=False, n_runs=1):
    """
    Compute rank-n_components SVD embeddings of arr.

    Parameters
    ----------
    arr: np.array of shape (n_samples, n_features)
        Data matrix
    n_components: int, default=20
        Number of components to keep
    randomized: bool, default=False
        Whether to use randomized SVD
    n_runs: int, default=1
        Run multiple times and take the realization with the lowest Frobenious reconstruction error

    Returns
    -------
    embeddings: array_like of shape (n_samples, n_components)
        Rank-n_comopnents SVD embedding of arr.
    """
    if n_components is None:
        return arr
    u, s, vh = robust_svd(arr, n_components=n_components, randomized=randomized, n_runs=n_runs)
    return u @ np.diag(s)



def center_scale(arr):
    """
    Column-wise center and scale by standard deviation.

    Parameters
    ----------
    arr: np.ndarray of shape (n_samples, n_features)

    Returns
    -------
    Center and scaled version of arr.
    """
    return (arr - arr.mean(axis=0)) / arr.std(axis=0)


def filter_bad_matches(matching, filter_prop=0.1):
    """
    Filter bad matches according to the distances of matched pairs.

    Parameters
    ----------
    matching: list
        rows, cols, vals = init_matching, where each matched pair is (rows[i], cols[i]),
        and their distance is vals[i]
    filter_prop: float
        Matched pairs with distance in top filter_prop are discarded
    Returns
    -------
    rows, cols, vals: list
        Each matched pair of rows[i], cols[i], their distance is vals[i]
    """
    init_rows, init_cols, init_vals = matching
    thresh = np.quantile(init_vals, 1 - filter_prop)
    rows = []
    cols = []
    vals = []
    for i, j, val in zip(init_rows, init_cols, init_vals):
        if val < thresh:
            rows.append(i)
            cols.append(j)
            vals.append(val)
    return np.array(rows, dtype=np.int32), np.array(cols, dtype=np.int32), np.array(vals, dtype=np.float32)

def pearson_correlation(arr1, arr2):
    """Calculate the vector of pearson correlations between each row of arr1 and arr2.

    Parameters
    ----------
    arr1: np.array of shape (n_samples, n_features)
        First dataset.
    arr2: np.array of shape (n_samples, n_features)
        Second dataset.

    Returns
    -------
    np.array of shape (n_samples,), the i-th entry is the pearson correlation between arr1[i, :] and arr2[i, :].
    """
    n, p = arr1.shape
    m, p2 = arr2.shape
    assert n == m and p2 == p

    arr1 = (arr1.T - np.mean(arr1, axis=1)).T
    arr2 = (arr2.T - np.mean(arr2, axis=1)).T

    arr1 = (arr1.T / np.sqrt(1e-6 + np.sum(arr1 ** 2, axis=1))).T
    arr2 = (arr2.T / np.sqrt(1e-6 + np.sum(arr2 ** 2, axis=1))).T

    return np.sum(arr1 * arr2, axis=1)

def cca_embedding(arr1, arr2, init_matching, filter_prop, n_components, max_iter=2000):
    """
    Filter bad matched pairs, align arr1 and arr2 using init_matching, fit CCA, and get CCA embeddings of arr1 and arr2.

    Parameters
    ----------
    arr1: np.ndarray of shape (n_samples1, n_features1)
        The first data matrix
    arr2: np.ndarray of shape (n_samples2, n_features2)
        The second data matrix
    init_matching: list
        rows, cols, vals = init_matching, where each matched pair is (rows[i], cols[i]),
        and their distance is vals[i]
    filter_prop: float
        Matched pairs with distance in top filter_prop are discarded when fitting CCA
    n_components: int
        Number of components to keep when fitting CCA
    max_iter: int, default=2000
        Maximum number of iterations for CCA

    Returns
    -------
    arr1_cca: np.array of shape (n_samples1, n_components)
    arr2_cca: np.array of shape (n_samples2, n_components)
    canonical_correlations: np.array of shape (n_components,)
    """

    # filter bad matched pairs
    arr1_indices, arr2_indices, _ = filter_bad_matches(init_matching, filter_prop)

    # align
    arr1_aligned = arr1[arr1_indices, :]
    arr2_aligned = arr2[arr2_indices, :]

    # cca
    cca = CCA(n_components=n_components, max_iter=max_iter)
    cca.fit(arr1_aligned, arr2_aligned)
    arr1_aligned_cca, arr2_aligned_cca = cca.transform(arr1_aligned, arr2_aligned)
    arr1_aligned_cca = center_scale(arr1_aligned_cca)
    arr2_aligned_cca = center_scale(arr2_aligned_cca)

    canonical_correlations = np.corrcoef(
        arr1_aligned_cca, arr2_aligned_cca, rowvar=False).diagonal(offset=n_components)
    arr1_cca, arr2_cca = cca.transform(arr1, arr2)
    arr1_cca = center_scale(arr1_cca)
    arr2_cca = center_scale(arr2_cca)

    return arr1_cca, arr2_cca, canonical_correlations

def clr_normalize_each_cell(adata, inplace=True):
    """Normalize count vector for each cell, i.e. for each row of .X"""

    def seurat_clr(x):
        """ A modified CLR function that supports sparse input """
        if scipy.sparse.issparse(x):
            # Only take non-zero entries for calculation
            nonzero_vals = x.data
            log_nonzero_vals = np.log1p(nonzero_vals)
            s = np.sum(log_nonzero_vals)
            cell_count = x.shape[1]
            exp = np.exp(s / cell_count)
            x.data = np.log1p(nonzero_vals / exp)
            return x
        else:
            # Dense array handling (as fallback)
            s = np.sum(np.log1p(x[x > 0]))
            exp = np.exp(s / len(x))
            return np.log1p(x / exp)

    if not inplace:
        adata = adata.copy()

    if scipy.sparse.issparse(adata.X):
        # Apply the CLR function to each row of the sparse matrix
        adata.X = scipy.sparse.vstack([seurat_clr(row) for row in adata.X])
    else:
        # Apply the CLR function to each row of the dense matrix
        adata.X = np.apply_along_axis(seurat_clr, 1, adata.X)

    return adata


def tfidf(X, n_components, binarize=True, random_state=0):
    import sklearn

    sc_count = np.copy(X)
    if binarize:
        sc_count = np.where(sc_count < 1, sc_count, 1)

    tfidf = TfidfTransformer(norm='l2', sublinear_tf=True)
    normed_count = tfidf.fit_transform(sc_count)

    lsi = sklearn.decomposition.TruncatedSVD(n_components=n_components, random_state=random_state)
    lsi_r = lsi.fit_transform(normed_count)

    X_lsi = lsi_r[:, 1:]

    return normed_count, X_lsi

def TFIDF_LSI(adata, n_comps=50, binarize=True, random_state=0):
    '''
    Computes LSI based on a TF-IDF transformation of the data from MultiMap. Putative dimensionality
    reduction for scATAC-seq data. Adds an ``.obsm['X_lsi']`` field to the object it was ran on.

    Input
    -----
    adata : ``AnnData``
        The object to run TFIDF + LSI on. Will use ``.X`` as the input data.
    n_comps : ``int``
        The number of components to generate. Default: 50
    binarize : ``bool``
        Whether to binarize the data prior to the computation. Often done during scATAC-seq
        processing. Default: True
    random_state : ``int``
        The seed to use for randon number generation. Default: 0
    '''

    # this is just a very basic wrapper for the non-adata function
    import scipy
    if scipy.sparse.issparse(adata.X):
        adata.X, adata.obsm['X_lsi'] = tfidf(adata.X.todense(), n_components=n_comps, binarize=binarize,
                                    random_state=random_state)
    else:
        adata.X, adata.obsm['X_lsi'] = tfidf(adata.X, n_components=n_comps, binarize=binarize, random_state=random_state)

## Predict gene scores based on chromatin accessibility
## issue https://www.biostars.org/p/114460/ 需要确保bedtools被安装了。
def _uniquify(seq, sep='-'):
    """Uniquify a list of strings.

    Adding unique numbers to duplicate values.

    Parameters
    ----------
    seq : `list` or `array-like`
        A list of values
    sep : `str`
        Separator

    Returns
    -------
    seq: `list` or `array-like`
        A list of updated values
    """

    dups = {}

    for i, val in enumerate(seq):
        if val not in dups:
            # Store index of first occurrence and occurrence value
            dups[val] = [i, 1]
        else:
            # Increment occurrence value, index value doesn't matter anymore
            dups[val][1] += 1

            # Use stored occurrence value
            seq[i] += (sep+str(dups[val][1]))

    return(seq)


class GeneScores:
    """A class used to represent gene scores

    Attributes
    ----------

    Methods
    -------

    """
    def __init__(self,
                 adata,
                 genome,
                 gene_anno=None,
                 tss_upstream=1e5,
                 tss_downsteam=1e5,
                 gb_upstream=5000,
                 cutoff_weight=1,
                 use_top_pcs=False,
                 use_precomputed=True,
                 use_gene_weigt=True,
                 min_w=1,
                 max_w=5):
        """
        Parameters
        ----------
        adata: `Anndata`
            Input anndata
        genome : `str`
            The genome name
        """
        self.adata = adata
        self.genome = genome
        self.gene_anno = gene_anno
        self.tss_upstream = tss_upstream
        self.tss_downsteam = tss_downsteam
        self.gb_upstream = gb_upstream
        self.cutoff_weight = cutoff_weight
        self.use_top_pcs = use_top_pcs
        self.use_precomputed = use_precomputed
        self.use_gene_weigt = use_gene_weigt
        self.min_w = min_w
        self.max_w = max_w

    def _read_gene_anno(self):
        """Read in gene annotation

        Parameters
        ----------

        Returns
        -------

        """
        assert (self.genome in ['hg19', 'hg38', 'mm9', 'mm10']),\
            "`genome` must be one of ['hg19','hg38','mm9','mm10']"

        bin_str = pkgutil.get_data('Garfield',
                                   f'data/gene_anno/{self.genome}_genes.bed')
        gene_anno = pd.read_csv(io.BytesIO(bin_str),
                                encoding='utf8',
                                sep='\t',
                                header=None,
                                names=['chr', 'start', 'end',
                                       'symbol', 'strand'])
        self.gene_anno = gene_anno
        return self.gene_anno

    def _extend_tss(self, pbt_gene):
        """Extend transcription start site in both directions

        Parameters
        ----------

        Returns
        -------

        """
        ext_tss = pbt_gene
        if(ext_tss['strand'] == '+'):
            ext_tss.start = max(0, ext_tss.start - self.tss_upstream)
            ext_tss.end = max(ext_tss.end, ext_tss.start + self.tss_downsteam)
        else:
            ext_tss.start = max(0, min(ext_tss.start,
                                       ext_tss.end - self.tss_downsteam))
            ext_tss.end = ext_tss.end + self.tss_upstream
        return ext_tss

    def _extend_genebody(self, pbt_gene):
        """Extend gene body upstream

        Parameters
        ----------

        Returns
        -------

        """
        ext_gb = pbt_gene
        if(ext_gb['strand'] == '+'):
            ext_gb.start = max(0, ext_gb.start - self.gb_upstream)
        else:
            ext_gb.end = ext_gb.end + self.gb_upstream
        return ext_gb

    def _weight_genes(self):
        """Weight genes

        Parameters
        ----------

        Returns
        -------

        """
        gene_anno = self.gene_anno
        gene_size = gene_anno['end'] - gene_anno['start']
        w = 1/gene_size
        w_scaled = (self.max_w-self.min_w) * (w-min(w)) / (max(w)-min(w)) \
            + self.min_w
        return w_scaled

    def cal_gene_scores(self):
        """Calculate gene scores

        Parameters
        ----------

        Returns
        -------

        """
        adata = self.adata
        if self.gene_anno is None:
            gene_ann = self._read_gene_anno()
        else:
            gene_ann = self.gene_anno

        df_gene_ann = gene_ann.copy()
        df_gene_ann.index = _uniquify(df_gene_ann['symbol'].values)
        if self.use_top_pcs:
            mask_p = adata.var['top_pcs']
        else:
            mask_p = pd.Series(True, index=adata.var_names)
        df_peaks = adata.var[mask_p][['chr', 'start', 'end']].copy()

        if('gene_scores' not in adata.uns_keys()):
            print('Gene scores are being calculated for the first time')
            print('`use_precomputed` has been ignored')
            self.use_precomputed = False

        if(self.use_precomputed):
            print('Using precomputed overlap')
            df_overlap_updated = adata.uns['gene_scores']['overlap'].copy()
        else:
            # add the fifth column
            # so that pybedtool can recognize the sixth column as the strand
            df_gene_ann_for_pbt = df_gene_ann.copy()
            df_gene_ann_for_pbt['score'] = 0
            df_gene_ann_for_pbt = df_gene_ann_for_pbt[['chr', 'start', 'end',
                                                       'symbol', 'score',
                                                       'strand']]
            df_gene_ann_for_pbt['id'] = range(df_gene_ann_for_pbt.shape[0])

            df_peaks_for_pbt = df_peaks.copy()
            df_peaks_for_pbt['id'] = range(df_peaks_for_pbt.shape[0])

            pbt_gene_ann = pybedtools.BedTool.from_dataframe(
                df_gene_ann_for_pbt
                )
            pbt_gene_ann_ext = pbt_gene_ann.each(self._extend_tss)
            pbt_gene_gb_ext = pbt_gene_ann.each(self._extend_genebody)

            pbt_peaks = pybedtools.BedTool.from_dataframe(df_peaks_for_pbt)

            # peaks overlapping with extended TSS
            pbt_overlap = pbt_peaks.intersect(pbt_gene_ann_ext,
                                              wa=True,
                                              wb=True)
            df_overlap = pbt_overlap.to_dataframe(
                names=[x+'_p' for x in df_peaks_for_pbt.columns]
                + [x+'_g' for x in df_gene_ann_for_pbt.columns])
            # peaks overlapping with gene body
            pbt_overlap2 = pbt_peaks.intersect(pbt_gene_gb_ext,
                                               wa=True,
                                               wb=True)
            df_overlap2 = pbt_overlap2.to_dataframe(
                names=[x+'_p' for x in df_peaks_for_pbt.columns]
                + [x+'_g' for x in df_gene_ann_for_pbt.columns])

            # add distance and weight for each overlap
            df_overlap_updated = df_overlap.copy()
            df_overlap_updated['dist'] = 0

            for i, x in enumerate(df_overlap['symbol_g'].unique()):
                # peaks within the extended TSS
                df_overlap_x = \
                    df_overlap[df_overlap['symbol_g'] == x].copy()
                # peaks within the gene body
                df_overlap2_x = \
                    df_overlap2[df_overlap2['symbol_g'] == x].copy()
                # peaks that are not intersecting with the promoter
                # and gene body of gene x
                id_overlap = df_overlap_x.index[
                    ~np.isin(df_overlap_x['id_p'], df_overlap2_x['id_p'])]
                mask_x = (df_gene_ann['symbol'] == x)
                range_x = df_gene_ann[mask_x][['start', 'end']].values\
                    .flatten()
                if(df_overlap_x['strand_g'].iloc[0] == '+'):
                    df_overlap_updated.loc[id_overlap, 'dist'] = pd.concat(
                        [abs(df_overlap_x.loc[id_overlap, 'start_p']
                             - (range_x[1])),
                         abs(df_overlap_x.loc[id_overlap, 'end_p']
                             - max(0, range_x[0]-self.gb_upstream))],
                        axis=1, sort=False).min(axis=1)
                else:
                    df_overlap_updated.loc[id_overlap, 'dist'] = pd.concat(
                        [abs(df_overlap_x.loc[id_overlap, 'start_p']
                             - (range_x[1]+self.gb_upstream)),
                         abs(df_overlap_x.loc[id_overlap, 'end_p']
                             - (range_x[0]))],
                        axis=1, sort=False).min(axis=1)

                n_batch = int(df_gene_ann_for_pbt.shape[0]/5)
                if(i % n_batch == 0):
                    print(f'Processing: {i/df_gene_ann_for_pbt.shape[0]:.1%}')
            df_overlap_updated['dist'] = df_overlap_updated['dist']\
                .astype(float)

            adata.uns['gene_scores'] = dict()
            adata.uns['gene_scores']['overlap'] = df_overlap_updated.copy()

        df_overlap_updated['weight'] = np.exp(
            -(df_overlap_updated['dist'].values/self.gb_upstream))
        mask_w = (df_overlap_updated['weight'] < self.cutoff_weight)
        df_overlap_updated.loc[mask_w, 'weight'] = 0
        # construct genes-by-peaks matrix
        mat_GP = csr_matrix(coo_matrix((df_overlap_updated['weight'],
                                       (df_overlap_updated['id_g'],
                                        df_overlap_updated['id_p'])),
                                       shape=(df_gene_ann.shape[0],
                                              df_peaks.shape[0])))
        # adata_GP = ad.AnnData(X=csr_matrix(mat_GP),
        #                       obs=df_gene_ann,
        #                       var=df_peaks)
        # adata_GP.layers['weight'] = adata_GP.X.copy()
        if self.use_gene_weigt:
            gene_weights = self._weight_genes()
            gene_scores = adata[:, mask_p].X * \
                (mat_GP.T.multiply(gene_weights))
        else:
            gene_scores = adata[:, mask_p].X * mat_GP.T
        adata_CG_atac = ad.AnnData(gene_scores,
                                   obs=adata.obs.copy(),
                                   var=df_gene_ann.copy())
        return adata_CG_atac


def gene_scores(adata,
                genome,
                gene_anno=None,
                tss_upstream=1e5,
                tss_downsteam=1e5,
                gb_upstream=5000,
                cutoff_weight=1,
                use_top_pcs=True,
                use_precomputed=True,
                use_gene_weigt=True,
                min_w=1,
                max_w=5):
    """Calculate gene scores

    Parameters
    ----------
    adata : AnnData
        Annotated data matrix.
    genome : `str`
        Reference genome. Choose from {'hg19', 'hg38', 'mm9', 'mm10'}
    gene_anno : `pandas.DataFrame`, optional (default: None)
        Dataframe of gene annotation.
        If None, built-in gene annotation will be used depending on `genome`;
        If provided, custom gene annotation will be used instead.
    tss_upstream : `int`, optional (default: 1e5)
        The number of base pairs upstream of TSS
    tss_downsteam : `int`, optional (default: 1e5)
        The number of base pairs downstream of TSS
    gb_upstream : `int`, optional (default: 5000)
        The number of base pairs upstream by which gene body is extended.
        Peaks within the extended gene body are given the weight of 1.
    cutoff_weight : `float`, optional (default: 1)
        Weight cutoff for peaks
    use_top_pcs : `bool`, optional (default: True)
        If True, only peaks associated with top PCs will be used
    use_precomputed : `bool`, optional (default: True)
        If True, overlap bewteen peaks and genes
        (stored in `adata.uns['gene_scores']['overlap']`) will be imported
    use_gene_weigt : `bool`, optional (default: True)
        If True, for each gene, the number of peaks assigned to it
        will be rescaled based on gene size
    min_w : `int`, optional (default: 1)
        The minimum weight for each gene.
        Only valid if `use_gene_weigt` is True
    max_w : `int`, optional (default: 5)
        The maximum weight for each gene.
        Only valid if `use_gene_weigt` is True

    Returns
    -------
    adata_new: AnnData
        Annotated data matrix.
        Stores #cells x #genes gene score matrix

    updates `adata` with the following fields.
    overlap: `pandas.DataFrame`, (`adata.uns['gene_scores']['overlap']`)
        Dataframe of overlap between peaks and genes
    """
    GS = GeneScores(adata,
                    genome,
                    gene_anno=gene_anno,
                    tss_upstream=tss_upstream,
                    tss_downsteam=tss_downsteam,
                    gb_upstream=gb_upstream,
                    cutoff_weight=cutoff_weight,
                    use_top_pcs=use_top_pcs,
                    use_precomputed=use_precomputed,
                    use_gene_weigt=use_gene_weigt,
                    min_w=min_w,
                    max_w=max_w)
    adata_CG_atac = GS.cal_gene_scores()
    return adata_CG_atac




# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from datetime import datetime
sys.path.insert(0, os.path.abspath('../Garfield'))
sys.path.insert(0, os.path.abspath('_ext'))
import Garfield  # noqa: E402


# -- Project information -----------------------------------------------------

project = 'Garfield'
copyright = '2024, Weige Zhou'
author = 'Weige Zhou'
copyright = f'{datetime.now():%Y}, {author}.'

# The full version, including alpha/beta/rc tags
release = Garfield.__version__


# -- Retrieve notebooks (borrowed from scVelo) -------------------------------

# from urllib.request import urlretrieve  # noqa: E402
#
# notebooks_url = "https://github.com/zhou-1314/Garfield_tutorials/tree/main/"
# notebooks_v1_0 = [
#     "01.panc8_all_batch_correction_test.ipynb",
#     # "02.panc_mapping.ipynb",
#     # "03.10x_pbmc_paired_scMulti_analysis.ipynb",
#     # "04.Garfield_spatia_molecular_niche_seqFISH_Mouse_Organogenesis.ipynb",
#     # "05.Garfield_spatial_niche_slideseqv2_mouse_hippocampus.ipynb",
#     # "06.Garfield_spMulti_mouse_brain.ipynb",
# ]
# notebooks_v1_1 = [
#     # "rna_10x_mouse_brain_1p3M.ipynb",
# ]
# for nb in notebooks_v1_0:
#     try:
#         urlretrieve(notebooks_url + "v1.0/" + nb, nb)
#     except Exception:
#         pass
#
# for nb in notebooks_v1_1:
#     try:
#         urlretrieve(notebooks_url + "v1.1/" + nb, nb)
#     except Exception:
#         pass

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.

needs_sphinx = "3.0"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    'sphinx.ext.napoleon',
    'sphinx.ext.doctest',
    'sphinx.ext.coverage',
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "nbsphinx",
    "edit_on_github",
    ]

# Generate the API documentation when building
autosummary_generate = True
autodoc_member_order = 'bysource'

# Napoleon settings
napoleon_google_docstring = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    "navigation_depth": 1,
    "titles_only": True,
    'logo_only': True,
}
html_context = dict(
    github_user='zhou-1314',  # Username
    github_repo='Garfield',  # Repo name
    github_version='main',  # Version
    conf_py_path='/docs/',  # Path in the checkout to the docs root
)
html_show_sphinx = False
html_logo = '_static/img/logo_garfield.png'
html_favicon = '_static/img/garfield_icon.svg'
# github_repo = 'Garfield'
# github_nb_repo = 'Garfield_tutorials'


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".

html_static_path = ['_static']



"""
This module contains dataloaders for the training of Garfield model.
"""

from typing import Optional

from torch_geometric.data import Data
from torch_geometric.loader import LinkNeighborLoader, NeighborLoader


def initialize_dataloaders(node_masked_data: Data,
                           edge_train_data: Optional[Data] = None,
                           edge_val_data: Optional[Data] = None,
                           edge_batch_size: Optional[int] = 64,
                           node_batch_size: int = 64,
                           n_direct_neighbors: int = -1,
                           n_hops: int = 1,
                           shuffle: bool = True,
                           edges_directed: bool = False,
                           neg_edge_sampling_ratio: float = 1.) -> dict:
    """
    Initialize edge-level and node-level training and validation dataloaders.

    Parameters
    ----------
    node_masked_data:
        PyG Data object with node-level split masks.
    edge_train_data:
        PyG Data object containing the edge-level training set.
    edge_val_data:
        PyG Data object containing the edge-level validation set.
    edge_batch_size:
        Batch size for the edge-level dataloaders.
    node_batch_size:
        Batch size for the node-level dataloaders.
    n_direct_neighbors:
        Number of sampled direct neighbors of the current batch nodes to be
        included in the batch. Defaults to ´-1´, which means to include all
        direct neighbors.
    n_hops:
        Number of neighbor hops / levels for neighbor sampling of nodes to be
        included in the current batch. E.g. ´2´ means to not only include
        sampled direct neighbors of current batch nodes but also sampled
        neighbors of the direct neighbors.
    shuffle:
        If `True`, shuffle the dataloaders.
    edges_directed:
        If `False`, both symmetric edge index pairs are included in the same
        edge-level batch (1 edge has 2 symmetric edge index pairs).
    neg_edge_sampling_ratio:
        Negative sampling ratio of edges. This is currently implemented in an
        approximate way, i.e. negative edges may contain false negatives.

    Returns
    ----------
    loader_dict:
        Dictionary containing training and validation PyG LinkNeighborLoader
        (for edge reconstruction) and NeighborLoader (for gene expression
        reconstruction) objects.
    """
    loader_dict = {}

    # Node-level dataloaders
    loader_dict["node_train_loader"] = NeighborLoader(
        node_masked_data,
        num_neighbors=[n_direct_neighbors] * n_hops,
        batch_size=node_batch_size,
        directed=False,
        shuffle=shuffle,
        input_nodes=node_masked_data.train_mask)
    if node_masked_data.val_mask.sum() != 0:
        loader_dict["node_val_loader"] = NeighborLoader(
            node_masked_data,
            num_neighbors=[n_direct_neighbors] * n_hops,
            batch_size=node_batch_size,
            directed=False,
            shuffle=shuffle,
            input_nodes=node_masked_data.val_mask)

    # Edge-level dataloaders
    if edge_train_data is not None:
        loader_dict["edge_train_loader"] = LinkNeighborLoader(
            edge_train_data,
            num_neighbors=[n_direct_neighbors] * n_hops,
            batch_size=edge_batch_size,
            edge_label=None,  # will automatically be added as 1 for all edges
            edge_label_index=edge_train_data.edge_label_index[:, edge_train_data.edge_label.bool()],
            # limit the edges to the ones from the edge_label_adj
            directed=edges_directed,
            shuffle=shuffle,
            neg_sampling_ratio=neg_edge_sampling_ratio)
    if edge_val_data is not None and edge_val_data.edge_label.sum() != 0:
        loader_dict["edge_val_loader"] = LinkNeighborLoader(
            edge_val_data,
            num_neighbors=[n_direct_neighbors] * n_hops,
            batch_size=edge_batch_size,
            edge_label=None,  # will automatically be added as 1 for all edges
            edge_label_index=edge_val_data.edge_label_index[:, edge_val_data.edge_label.bool()],
            # limit the edges to the ones from the edge_label_adj
            directed=edges_directed,
            shuffle=shuffle,
            neg_sampling_ratio=neg_edge_sampling_ratio)

    return loader_dict


"""
This module contains metrics to evaluate the Garfield model training.
"""

from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import sklearn.metrics as skm
import torch
from matplotlib.ticker import MaxNLocator


def eval_metrics(
        edge_recon_probs: Union[torch.Tensor, np.ndarray],
        edge_labels: Union[torch.Tensor, np.ndarray],
        omics_recon_pred: Union[torch.Tensor, np.ndarray]=None,
        omics_recon_truth: Union[torch.Tensor, np.ndarray]=None) -> dict:
    """
    Get the evaluation metrics for a (balanced) sample of positive and negative
    edges and a sample of nodes.

    Parameters
    ----------
    edge_recon_probs:
        Tensor or array containing reconstructed edge probabilities.
    edge_labels:
        Tensor or array containing ground truth labels of edges.

    Returns
    ----------
    eval_dict:
        Dictionary containing the evaluation metrics ´auroc_score´ (area under
        the  receiver operating characteristic curve), ´auprc score´ (area under
        the precision-recall curve), ´best_acc_score´ (accuracy under optimal
        classification threshold) and ´best_f1_score´ (F1 score under optimal
        classification threshold).
    """
    eval_dict = {}

    if isinstance(edge_recon_probs, torch.Tensor):
        edge_recon_probs = edge_recon_probs.detach().cpu().numpy()
    if isinstance(edge_labels, torch.Tensor):
        edge_labels = edge_labels.detach().cpu().numpy()
    if isinstance(omics_recon_pred, torch.Tensor):
        omics_recon_pred = omics_recon_pred.detach().cpu().numpy()
    if isinstance(omics_recon_truth, torch.Tensor):
        omics_recon_truth = omics_recon_truth.detach().cpu().numpy()

    if omics_recon_pred is not None and omics_recon_truth is not None:
        # Calculate the gene expression mean squared error
        eval_dict["gene_expr_mse_score"] = skm.mean_squared_error(
            omics_recon_truth,
            omics_recon_pred)

    # Calculate threshold independent metrics
    eval_dict["auroc_score"] = skm.roc_auc_score(edge_labels, edge_recon_probs)
    eval_dict["auprc_score"] = skm.average_precision_score(edge_labels,
                                                           edge_recon_probs)

    # Get the optimal classification probability threshold above which an edge
    # is classified as positive so that the threshold optimizes the accuracy
    # over the sampled (balanced) set of positive and negative edges.
    best_acc_score = 0
    best_threshold = 0
    for threshold in np.arange(0.01, 1, 0.005):
        pred_labels = (edge_recon_probs > threshold).astype("int")
        acc_score = skm.accuracy_score(edge_labels, pred_labels)
        if acc_score > best_acc_score:
            best_threshold = threshold
            best_acc_score = acc_score
    eval_dict["best_acc_score"] = best_acc_score
    eval_dict["best_acc_threshold"] = best_threshold

    # Get the optimal classification probability threshold above which an edge
    # is classified as positive so that the threshold optimizes the F1 score
    # over the sampled (balanced) set of positive and negative edges.
    best_f1_score = 0
    for threshold in np.arange(0.01, 1, 0.005):
        pred_labels = (edge_recon_probs > threshold).astype("int")
        f1_score = skm.f1_score(edge_labels, pred_labels)
        if f1_score > best_f1_score:
            best_f1_score = f1_score
    eval_dict["best_f1_score"] = best_f1_score
    return eval_dict


def plot_eval_metrics(eval_dict: dict) -> plt.figure:
    """
    Plot evaluation metrics.

    Parameters
    ----------
    eval_dict:
        Dictionary containing the eval metric scores to be plotted.

    Returns
    ----------
    fig:
        Matplotlib figure containing a plot of the evaluation metrics.
    """
    # Plot epochs as integers
    ax = plt.figure().gca()
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))

    # Plot eval metrics
    for metric_key, metric_scores in eval_dict.items():
        plt.plot(metric_scores, label=metric_key)
    plt.title("Evaluation metrics over epochs")
    plt.ylabel("metric score")
    plt.xlabel("epoch")
    plt.legend(loc="lower right")

    # Retrieve figure
    fig = plt.gcf()
    plt.close()
    return fig


from anndata import AnnData
from mudata import MuData

# read data
from ..data.datareaders import concat_data
# preprocessing
from ..preprocessing.preprocess_utils import preprocessing


def DataProcess(
         adata_list,
         profile,
         data_type=None,
         sub_data_type=None,
         sample_col='batch',
         genome=None,
         weight=None,
         graph_const_method=None,
         use_gene_weigt=True,
         use_top_pcs=False,
         used_hvg=True,
         min_features=100,
         min_cells=3,
         keep_mt=False,
         target_sum=1e4,
         rna_n_top_features=3000,
         atac_n_top_features=10000,
         n_components=50,
         n_neighbors=15,
         metric='correlation',
         svd_solver='arpack'):
    """
    Processes single or multi-modal data (e.g., RNA, ATAC, ADT, spatial) with optional preprocessing steps
    such as normalization, feature selection, and dimensionality reduction.

    Parameters
    ----------
    adata_list : list of AnnData or MuData objects
        List of AnnData or MuData objects to be concatenated and processed.
    profile : str
        Data profile type, e.g., 'RNA', 'ATAC', 'ADT', 'multi-modal', or 'spatial'.
    data_type : str, optional
        Type of data being processed, e.g., 'single-cell', 'bulk'. Default is None.
    sub_data_type : list[str], optional
        List of sub-data types for multi-modal data, e.g., ['rna', 'atac'] or ['rna', 'adt']. Default is None.
    sample_col : str, optional
        Column in the dataset used to indicate batch or sample groupings. Default is 'batch'.
    genome : str, optional
        Reference genome for the dataset. Default is None.
    weight : float or None, optional
        Weight for certain data processing steps, such as graph construction. Default is None.
    graph_const_method : str, optional
        Method for constructing the graph if applicable, e.g., 'knn'. Default is None.
    use_gene_weigt : bool, optional
        Whether to use gene weights in the preprocessing steps. Default is True.
    use_top_pcs : bool, optional
        Whether to use the top principal components during dimensionality reduction. Default is False.
    used_hvg : bool, optional
        Whether to use highly variable genes (HVG) for the analysis. Default is True.
    min_features : int, optional
        Minimum number of features required for a cell to be included. Default is 100.
    min_cells : int, optional
        Minimum number of cells required for a feature to be included. Default is 3.
    keep_mt : bool, optional
        Whether to keep mitochondrial genes in the dataset. Default is False.
    target_sum : float, optional
        Target sum for normalization. Default is 1e4.
    rna_n_top_features : int, optional
        Number of top features to keep for RNA data. Default is 3000.
    atac_n_top_features : int, optional
        Number of top features to keep for ATAC data. Default is 10000.
    n_components : int, optional
        Number of components for dimensionality reduction (e.g., PCA). Default is 50.
    n_neighbors : int, optional
        Number of neighbors for graph-based algorithms. Default is 15.
    metric : str, optional
        Distance metric to use in graph construction. Default is 'correlation'.
    svd_solver : str, optional
        Solver to use for singular value decomposition (SVD). Default is 'arpack'.

    Returns
    -------
    AnnData or MuData
        Preprocessed single or multi-modal data based on the specified profile and sub_data_type.
    """
    # load data
    adata = concat_data(
        adata_list,
        batch_categories=None,
        join='inner',
        batch_key=sample_col,  # 'batch'
        index_unique=None,
        save=None
    )
    if isinstance(adata, AnnData):
        if adata.X.max() < 50:
            print('Warning: adata.X may have already been normalized, adata.X must be `counts`, please check.')
        else:
            adata.layers['counts'] = adata.X.copy()
    elif isinstance(adata, MuData):
        if adata.mod['rna'].X.max() < 50:
            print('Warning: adata.X may have already been normalized, adata.X must be `counts`, please check.')
        else:
            adata.mod['rna'].layers['counts'] = adata.mod['rna'].X.copy()

    # RNA ATAC ADT
    if profile in ['RNA', 'ATAC', 'ADT']:

        ## 预处理
        _, adata_hvg = preprocessing(
            adata,
            profile=profile,
            data_type=data_type,
            genome=genome,
            use_gene_weigt=use_gene_weigt,
            use_top_pcs=use_top_pcs,
            used_hvgs=used_hvg,
            min_features=min_features,
            min_cells=min_cells,
            target_sum=target_sum,
            rna_n_top_features=rna_n_top_features,
            atac_n_top_features=atac_n_top_features,
            n_components=n_components,
            n=n_neighbors,
            batch_key=sample_col,
            metric=metric,
            svd_solver=svd_solver,
            keep_mt=keep_mt
        )

        return adata_hvg

    ### Paired multi-modal
    elif profile == 'multi-modal':
        if len(sub_data_type) == 2:
            if sub_data_type[0] == 'rna' and sub_data_type[1] == 'atac':
                rna_adata = adata.mod['rna'].copy()
                atac_adata = adata.mod['atac'].copy()
                mdata = MuData({"rna": rna_adata, "atac": atac_adata})
            elif sub_data_type[0] == 'rna' and sub_data_type[1] == 'adt':
                rna_adata = adata.mod['rna'].copy()
                adt_adata = adata.mod['adt'].copy()
                mdata = MuData({"rna": rna_adata, "adt": adt_adata})
        else:
            ValueError('The length of sub_data_type must be 2, such as: ["rna", "atac"] or ["rna", "adt"].')
        del adata

        ## 预处理
        merged_adata = preprocessing(
            mdata,
            profile=profile,
            data_type=data_type,
            sub_data_type=sub_data_type,
            genome=genome,
            weight=weight,
            use_gene_weigt=use_gene_weigt,
            use_top_pcs=use_top_pcs,
            used_hvgs=used_hvg,
            min_features=min_features,
            min_cells=min_cells,
            target_sum=target_sum,
            rna_n_top_features=rna_n_top_features,
            atac_n_top_features=atac_n_top_features,
            n_components=n_components,
            n=n_neighbors,
            batch_key=sample_col,
            metric=metric,
            svd_solver=svd_solver,
            keep_mt=keep_mt
        )

        return merged_adata

    ### spatial single- or multi-modal
    elif profile == 'spatial':
        ## 预处理
        merged_adata = preprocessing(
            adata,
            profile=profile,
            data_type=data_type,
            sub_data_type=sub_data_type,
            genome=genome,
            weight=weight,
            graph_const_method=graph_const_method,
            use_gene_weigt=use_gene_weigt,
            use_top_pcs=use_top_pcs,
            used_hvgs=used_hvg,
            min_features=min_features,
            min_cells=min_cells,
            target_sum=target_sum,
            rna_n_top_features=rna_n_top_features,
            atac_n_top_features=atac_n_top_features,
            n_components=n_components,
            n=n_neighbors,
            batch_key=sample_col,
            metric=metric,
            svd_solver=svd_solver,
            keep_mt=keep_mt
        )

        return merged_adata

    else:
        return "Unknown input data type."


"""
This module contains data processors for the training of Garfield model.
"""

from typing import List, Optional, Tuple

import scipy.sparse as sp
import torch
from anndata import AnnData
from torch_geometric.data import Data
from torch_geometric.transforms import RandomNodeSplit, RandomLinkSplit
from torch_geometric.utils import add_self_loops, remove_self_loops

from .datasets import GraphAnnTorchDataset


def edge_level_split(data: Data,
                     edge_label_adj: Optional[sp.csr_matrix],
                     val_ratio: float = 0.1,
                     test_ratio: float = 0.,
                     is_undirected: bool = True,
                     neg_sampling_ratio: float = 0.) -> Tuple[Data, Data, Data]:
    """
    Split a PyG Data object into training, validation and test PyG Data objects
    using an edge-level split. The training split does not include edges in the
    validation and test splits and the validation split does not include edges
    in the test split. However, nodes will not be split and all node features
    will be accessible from all splits.

    Check https://github.com/pyg-team/pytorch_geometric/issues/3668 for more
    context how RandomLinkSplit works.

    Parameters
    ----------
    data:
        PyG Data object to be split.
    edge_label_adj:
        Adjacency matrix which contains edges for edge reconstruction. If
        ´None´, uses the 'normal' adjacency matrix used for message passing.
    val_ratio:
        Ratio of edges to be included in the validation split.
    test_ratio:
        Ratio of edges to be included in the test split.
    is_undirected:
        If ´True´, the graph is assumed to be undirected, and positive and
        negative samples will not leak (reverse) edge connectivity across
        different splits. This is set to ´False´, as there is an issue with
        replication of self loops.
    neg_sampling_ratio:
        Ratio of negative sampling. This should be set to 0 if negative sampling
        is done by the dataloader.

    Returns
    ----------
    train_data:
        Training PyG Data object.
    val_data:
        Validation PyG Data object.
    test_data:
        Test PyG Data object.
    """
    # Clone data to not modify in place to not affect node split
    data_no_self_loops = data.clone()

    # Remove self loops temporarily as we don't want them as edge labels. There
    # is also an issue with RandomLinkSplit (self loops will be replicated for
    # message passing). We will add the self loops again after the split
    data_no_self_loops.edge_index, data_no_self_loops.edge_attr = (
        remove_self_loops(edge_index=data.edge_index,
                          edge_attr=data.edge_attr))

    if edge_label_adj is not None:
        # Add edge label which is 1 for edges from edge_label_adj and 0 otherwise.
        # This will be used by dataloader to only sample edges from edge_label_adj
        # as opposed to from adj.
        data_no_self_loops.edge_label = torch.tensor(
            [(edge_label_adj[edge_index[0].item(), edge_index[1].item()] == 1.0) for
             edge_index in data_no_self_loops.edge_attr]).int()

    random_link_split = RandomLinkSplit(
        num_val=val_ratio,
        num_test=test_ratio,
        is_undirected=is_undirected,
        key="edge_label",  # if ´edge_label´ is not existent, it will be added with 1s
        neg_sampling_ratio=neg_sampling_ratio)
    train_data, val_data, test_data = random_link_split(data_no_self_loops)

    # Readd self loops for message passing
    for split_data in [train_data, val_data, test_data]:
        split_data.edge_index = add_self_loops(
            edge_index=split_data.edge_index,
            num_nodes=split_data.x.shape[0])[0]
        split_data.edge_attr = add_self_loops(
            edge_index=split_data.edge_attr.t(),
            num_nodes=split_data.x.shape[0])[0].t()
    return train_data, val_data, test_data


def node_level_split_mask(data: Data,
                          val_ratio: float = 0.1,
                          test_ratio: float = 0.,
                          split_key: str = "x") -> Data:
    """
    Split data on node-level into training, validation and test sets by adding
    node-level masks (train_mask, val_mask, test_mask) to the PyG Data object.

    Parameters
    ----------
    data:
        PyG Data object to be split.
    val_ratio:
        Ratio of nodes to be included in the validation split.
    test_ratio:
        Ratio of nodes to be included in the test split.
    split_key:
        The attribute key of the PyG Data object that holds the ground
        truth labels. Only nodes in which the key is present will be split.

    Returns
    ----------
    data:
        PyG Data object with ´train_mask´, ´val_mask´ and ´test_mask´ attributes
        added.
    """
    random_node_split = RandomNodeSplit(
        num_val=val_ratio,
        num_test=test_ratio,
        key=split_key)
    data = random_node_split(data)
    return data


def prepare_data(adata,
                 label_name: str = None,
                 used_pca_feat: bool = False,
                 adj_key: str = "connectivities",
                 edge_label_adj_key: str = "edge_label_spatial_connectivities",
                 edge_val_ratio: float = 0.1,
                 edge_test_ratio: float = 0.,
                 node_val_ratio: float = 0.1,
                 node_test_ratio: float = 0.) -> dict:
    """
    This function performs node-level and edge-level splits and returns a dictionary containing the processed data.

    Parameters
    ----------
    adata : AnnData
        An AnnData object containing gene expression or other relevant data.
    label_name : str, optional
        The name of the label to use for node classification or regression tasks. Default is None.
    adj_key : str, optional
        Key in the AnnData object that corresponds to the adjacency matrix for graph construction. Default is "connectivities".
    edge_label_adj_key : str, optional
        Key in the AnnData object that corresponds to the adjacency matrix used for edge-label reconstruction tasks. Default is "edge_label_spatial_connectivities".
    edge_val_ratio : float, optional
        Proportion of edges to use for validation in the edge-level split. Default is 0.1.
    edge_test_ratio : float, optional
        Proportion of edges to use for testing in the edge-level split. Default is 0.
    node_val_ratio : float, optional
        Proportion of nodes to use for validation in the node-level split. Default is 0.1.
    node_test_ratio : float, optional
        Proportion of nodes to use for testing in the node-level split. Default is 0.

    Returns
    ----------
    dict
        A dictionary containing the following keys:
        - "edge_train_data": Training data for edge-level tasks.
        - "edge_val_data": Validation data for edge-level tasks.
        - "edge_test_data": Testing data for edge-level tasks.
        - "node_masked_data": Data with nodes masked for validation and testing in node-level tasks.
    """
    data_dict = {}
    dataset = GraphAnnTorchDataset(
        adata=adata,
        label_name=label_name,
        adj_key=adj_key,
        used_feat=used_pca_feat,
        edge_label_adj_key=edge_label_adj_key
    )
    # PyG Data object (has 2 edge index pairs for one edge because of symmetry;
    # one edge index pair will be removed in the edge-level split).
    # 确保两者的形状在连接时是匹配的
    assert dataset.edge_index.shape[1] == dataset.edge_weight.shape[0], \
        "Mismatch in the number of edges between edge_index and edge_weight"
    data_node = Data(x=dataset.x,
                y=dataset.y,
                edge_index=dataset.edge_index,
                edge_attr=torch.concat([dataset.edge_index.t(),
                                        dataset.edge_weight], dim=-1))  # store index of edge nodes as
    # edge attribute for aggregation weight retrieval in mini batches

    data_edge = Data(x=dataset.x,
                     y=dataset.y,
                     edge_index=dataset.edge_index,
                     edge_attr=dataset.edge_index.t())  # store index of edge nodes as
    # edge attribute for aggregation weight retrieval in mini batches

    # Edge-level split for edge reconstruction
    edge_train_data, edge_val_data, edge_test_data = edge_level_split(
        data=data_edge,
        edge_label_adj=dataset.edge_label_adj,
        val_ratio=edge_val_ratio,
        test_ratio=edge_test_ratio)
    data_dict["edge_train_data"] = edge_train_data
    data_dict["edge_val_data"] = edge_val_data
    data_dict["edge_test_data"] = edge_test_data

    # Node-level split for gene expression reconstruction
    data_dict["node_masked_data"] = node_level_split_mask(
        data=data_node,
        val_ratio=node_val_ratio,
        test_ratio=node_test_ratio)
    return data_dict


import sys

if sys.version_info < (3, 6):
    sys.exit('garfield requires Python >= 3.6')

from setuptools import setup, find_packages
from pathlib import Path

version = {}
with open("Garfield/_version.py") as fp:
    exec(fp.read(), version)

setup(
    name='Garfield',
    version=version['__version__'],
    author='Weige Zhou',
    author_email='zhouwg1314@gmail.com',
    license='BSD',
    description='Garfield: Graph-based Contrastive Learning enable Fast Single-Cell Embedding',
    long_description=Path('README.md').read_text('utf-8'),
    long_description_content_type="text/markdown",
    url='https://github.com/zhou-1314/Garfield',
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        x.strip() for x in
        Path('requirements.txt').read_text('utf-8').splitlines()
    ],
    include_package_data=True,
    package_data={"Garfield": ["data/gene_anno/*.bed"],
                  '': ['*.txt']}
)
